"""Converts directory tree containing Gherkin files into a tree which would be included into rst files

Usage:
    bdd_tree_to_rst.py <features_dir> <output_file>

"""
import sys
from collections import deque
from functools import reduce
from itertools import chain
from operator import methodcaller, truediv
from os.path import commonpath
from pathlib import Path
from textwrap import dedent

from docopt import docopt

SECTION_SYMBOLS = "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~"


def convert(features_path: Path, output_file_path: Path):
    base_output_common_path = Path(commonpath([str(features_path), str(output_file_path)]))
    features_path_rel_to_common_path = features_path.relative_to(base_output_common_path)
    output_path_rel_to_common_path = output_file_path.parent.relative_to(base_output_common_path)

    output_path_rel_to_features_path = (
        reduce(truediv, [".."] * len(output_path_rel_to_common_path.parts), Path()) / features_path_rel_to_common_path
    )
    processable_paths = deque([features_path])

    content = ""
    content += dedent(
        # language=rst
        """\
            .. NOTE:: Features below are part of end-to-end test suite; You always could find most specific
                      use cases of **pytest-bdd-ng** by investigation of its regression
                      test suite https://github.com/elchupanebrej/pytest-bdd-ng/tree/default/tests

        """
    )

    while processable_paths:
        processable_path = processable_paths.popleft()

        gherkin_file_paths = chain(processable_path.glob("*.gherkin"), processable_path.glob("*.feature"))
        struct_bdd_file_paths = processable_path.glob("*.bdd.yaml")

        sub_processable_paths = list(filter(methodcaller("is_dir"), processable_path.iterdir()))

        for path in gherkin_file_paths:
            content += dedent(
                # language=rst
                f"""\
                    {path.name}
                    {SECTION_SYMBOLS[len(path.relative_to(features_path).parts)-1]*len(path.name)}

                    .. include:: {(output_path_rel_to_features_path / path.relative_to(features_path)).as_posix()}
                       :code: gherkin

                """
            )

        for path in struct_bdd_file_paths:
            content += dedent(
                # language=rst
                f"""\
                    {path.name}
                    {SECTION_SYMBOLS[len(path.relative_to(features_path).parts)-1]*len(path.name)}

                    .. include:: {(output_path_rel_to_features_path / path.relative_to(features_path)).as_posix()}
                       :code: yaml

                """
            )

        processable_paths.extend(sub_processable_paths)
    return content.rstrip("\n") + "\n"


def main():  # pragma: no cover
    arguments = docopt(__doc__)
    features_dir = Path(arguments["<features_dir>"]).resolve()
    if not features_dir.exists() or not features_dir.is_dir():
        raise ValueError(f"Wrong input features directory {features_dir} is provided")
    output_file = Path(arguments["<output_file>"]).resolve()
    original_output = output_file.read_text() if output_file.exists() and output_file.is_file() else ""
    output = convert(features_dir, output_file)
    if output != original_output:
        output_file.parent.mkdir(parents=True, exist_ok=True)
        output_file.write_text(output)
        sys.exit("Documentation is generated and overwritten")


if __name__ == "__main__":  # pragma: no cover
    main()



from collections import defaultdict, namedtuple
from enum import Enum
from functools import partial
from inspect import getfile
from itertools import chain, product, starmap, zip_longest
from operator import attrgetter, eq, is_not
from pathlib import Path
from typing import Any, List, Optional, Sequence, Union

from attr import attrib, attrs
from pydantic import BaseModel, Extra, Field, validator

from pytest_bdd.compatibility.typing import Literal
from pytest_bdd.model.messages import KeywordType
from pytest_bdd.utils import deepattrgetter


class Keyword(Enum):
    Given = "Given"
    When = "When"
    Then = "Then"
    And = "And"
    But = "But"
    Star = "*"


class SubKeyword(Enum):
    Step = "Step"
    Alternative = "Alternative"


KEYWORD_TO_TYPE = defaultdict(
    lambda: KeywordType.unknown,
    [
        (Keyword.Given, KeywordType.context),
        (Keyword.When, KeywordType.action),
        (Keyword.Then, KeywordType.outcome),
        (Keyword.And, KeywordType.conjunction),
        (Keyword.But, KeywordType.conjunction),
        (Keyword.Star, KeywordType.unknown),
    ],
)


class Node(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    tags: Optional[Sequence[str]] = Field(default_factory=list, alias="Tags")
    name: Optional[str] = Field(None, alias="Name")
    description: Optional[str] = Field(None, alias="Description")
    comments: Optional[Sequence[str]] = Field(default_factory=list, alias="Comments")


class Table(Node):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    type: Optional[Literal["Rowed", "Columned"]] = Field("Rowed", alias="Type")
    parameters: Optional[Sequence[str]] = Field(default_factory=list, alias="Parameters")
    values: Optional[Sequence[Sequence[Any]]] = Field(default_factory=list, alias="Values")

    @property
    def columned_values(self):
        if self.type == "Columned":
            return self.values
        else:
            return list(zip(*self.values))

    @property
    def rowed_values(self):
        if self.type == "Rowed":
            return self.values
        else:
            return list(zip(*self.values))


class SubTable(Node):
    sub_table: Table = Field(..., alias="Table")


class Join(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    tables: List[Union[Table, "Join", SubTable]] = Field(default_factory=list, alias="Join")

    __hash__ = id

    @property
    def tags(self):
        return list(dict.fromkeys(chain.from_iterable(map(attrgetter("tags"), self.tables))))

    @property
    def name(self):
        return "\n".join(filter(partial(is_not, None), list(chain(map(attrgetter("name"), self.tables)))))

    @property
    def description(self):
        return "\n".join(
            filter(
                partial(is_not, None),
                chain.from_iterable(map(deepattrgetter("description", skip_missing=True), self.tables)),
            )
        )

    @property
    def comments(self):
        return list(chain.from_iterable(map(attrgetter("comments"), self.tables)))

    @property
    def parameters(self):
        return list(dict.fromkeys(chain.from_iterable(map(attrgetter("parameters"), self.tables))))

    @property
    def type(self):
        return "Rowed"

    @property
    def values(self):
        def _():
            filled_tables = list(filter(attrgetter("parameters"), self.tables))
            if filled_tables:
                filled_tables_parameters = list(chain.from_iterable(map(attrgetter("parameters"), self.tables)))
                for filled_tables_values in map(
                    lambda tables_values: list(chain.from_iterable(tables_values)),
                    product(*map(attrgetter("rowed_values"), filled_tables)),
                ):
                    if all(
                        [
                            all(
                                starmap(
                                    eq,
                                    product(
                                        [
                                            value
                                            for _parameter, value in zip(filled_tables_parameters, filled_tables_values)
                                            if parameter == _parameter
                                        ],
                                        repeat=2,
                                    ),
                                )
                            )
                            for parameter in self.parameters
                        ]
                    ):

                        def values_gen():
                            for parameter in self.parameters:
                                for _parameter, value in zip(filled_tables_parameters, filled_tables_values):
                                    if parameter == _parameter:
                                        yield value
                                        break

                        values = list(values_gen())
                        yield values
            else:
                yield from map(
                    lambda values_combination: list(chain.from_iterable(values_combination)),
                    product(*map(attrgetter("rowed_values"), self.tables)),
                )

        _values = list(_())
        return _values

    @property
    def columned_values(self):
        return list(zip(*self.values))

    @property
    def rowed_values(self):
        return self.values

    @validator("tables", each_item=True)
    def convert_sub_tables_to_tables(cls, value):
        if isinstance(value, SubTable):
            return value.sub_table
        else:
            return value


@validator("steps", pre=True, each_item=True)
def convert_to_step(cls, value):
    if isinstance(value, str):
        return Step(action=value)
    elif isinstance(value, dict) and len(value) == 1 and next(iter(value)) not in SubKeyword.__members__:
        return Step(type=next(iter(value.keys())), action=next(iter(value.values())))
    else:
        return value


@validator("type")
def select_step_keyword_type(cls, value, values, **kwargs):
    try:
        return Keyword(value)
    except:
        return value


@validator("steps", each_item=True)
def convert_sub_steps_to_steps(cls, value):
    if isinstance(value, SubStep):
        return value.sub_step
    else:
        return value


class StepPrototype(Node):
    steps: Sequence[
        Union["Alternative", "Step", "StarStep", "GivenStep", "WhenStep", "ThenStep", "AndStep", "ButStep", "SubStep"]
    ] = Field(default_factory=list, alias="Steps")

    type: Optional[Union[Keyword, str]] = Field(Keyword.Star, alias="Type")
    data: List[Union[Table, Join, SubTable]] = Field(default_factory=list, alias="Data")
    examples: List[Union[Table, Join, SubTable]] = Field(default_factory=list, alias="Examples")

    Route = namedtuple("Route", ["tags", "steps", "example_table"])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.keyword_type = KEYWORD_TO_TYPE[self.type]

    @property
    def routes(self):
        for routes in (
            product(*map(attrgetter("routes"), self.steps))
            if self.steps
            else [[self.Route([], [], Table(parameters=[], values=[]))]]
        ):
            steps = [self, *chain.from_iterable(map(attrgetter("steps"), routes))]

            if self.examples:
                for _example_table in self.examples:
                    example_table = Join(tables=[*map(attrgetter("example_table"), routes), _example_table])
                    tags = list(
                        {*chain.from_iterable(map(attrgetter("tags"), routes)), *example_table.tags, *self.tags}
                    )

                    yield self.Route(
                        tags,
                        steps,
                        example_table,
                    )
            else:
                example_table = Join(tables=[*map(attrgetter("example_table"), routes)])
                tags = list({*chain.from_iterable(map(attrgetter("tags"), routes)), *example_table.tags, *self.tags})

                yield self.Route(
                    tags,
                    steps,
                    example_table,
                )

    @classmethod
    def build_action_first_arg(cls, action, *args, **kwargs):
        return cls(*args, **kwargs, action=action)

    @validator("examples", each_item=True)
    def convert_sub_tables_to_tables_at_examples(cls, value):
        if isinstance(value, SubTable):
            return value.sub_table
        else:
            return value

    @validator("data", each_item=True)
    def convert_sub_tables_to_tables_at_data(cls, value):
        if isinstance(value, SubTable):
            return value.sub_table
        else:
            return value

    convert_sub_steps_to_steps = convert_sub_steps_to_steps

    @attrs
    class Locator:
        step: "StepPrototype" = attrib()
        filename = attrib()
        uri = attrib()

        def resolve(self, config):
            from pytest_bdd.struct_bdd.model_builder import GherkinDocumentBuilder

            feature = GherkinDocumentBuilder(self.step).build_feature(
                filename=self.filename, uri=self.uri, id_generator=config.pytest_bdd_id_generator
            )
            return zip_longest((), feature.pickles, fillvalue=feature)

    def as_test(self, filename):
        from pytest_bdd.scenario import scenarios

        return scenarios(
            locators=[self.Locator(self, str(Path(filename).as_posix()), str(Path(filename).relative_to(Path.cwd())))],
            return_test_decorator=False,
        )

    def as_test_decorator(self, filename):
        from pytest_bdd.scenario import scenarios

        return scenarios(
            locators=[self.Locator(self, str(Path(filename).as_posix()), str(Path(filename).relative_to(Path.cwd())))],
            return_test_decorator=True,
        )

    def __call__(self, func):
        return self.as_test_decorator(getfile(func))(func)


class Alternative(Node):
    steps: Sequence[
        Union["Alternative", "Step", "StarStep", "GivenStep", "WhenStep", "ThenStep", "AndStep", "ButStep", "SubStep"]
    ] = Field(default_factory=list, alias="Alternative")

    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    @property
    def routes(self):
        yield from chain.from_iterable(map(attrgetter("routes"), self.steps))

    @validator("steps", pre=True, each_item=True)
    def convert_step(cls, value):
        if isinstance(value, str):
            return Step(action=value)
        else:
            return value

    convert_sub_steps_to_steps = convert_sub_steps_to_steps


class Step(StepPrototype):
    steps: Sequence[
        Union["Alternative", "Step", "StarStep", "GivenStep", "WhenStep", "ThenStep", "AndStep", "ButStep", "SubStep"]
    ] = Field(default_factory=list, alias="Steps")
    type: Union[Keyword, Optional[str]] = Field(Keyword.Star, alias="Type")
    keyword_type: Optional[KeywordType] = Field(KeywordType.unknown)
    action: Optional[str] = Field(alias="Action")

    convert_to_step = convert_to_step
    select_step_keyword_type = select_step_keyword_type


class SubStep(BaseModel):
    sub_step: Step = Field(..., alias="Step")


class StarStep(StepPrototype):
    steps: Sequence[
        Union["Alternative", "Step", "StarStep", "GivenStep", "WhenStep", "ThenStep", "AndStep", "ButStep", "SubStep"]
    ] = Field(default_factory=list, alias="Steps")

    type: Optional[Keyword] = Field(Keyword.Star, alias="Type")
    keyword_type: Optional[KeywordType] = Field(KeywordType.unknown)
    action: Optional[str] = Field(alias=Keyword.Star.value)

    convert_to_step = convert_to_step
    select_step_keyword_type = select_step_keyword_type


class GivenStep(StepPrototype):
    steps: Sequence[
        Union["Alternative", "Step", "StarStep", "GivenStep", "WhenStep", "ThenStep", "AndStep", "ButStep", "SubStep"]
    ] = Field(default_factory=list, alias="Steps")
    type: Optional[Keyword] = Field(Keyword.Given)
    keyword_type: Optional[KeywordType] = Field(KeywordType.context, alias="Type")
    action: Optional[str] = Field(alias=Keyword.Given.value)

    convert_to_step = convert_to_step
    select_step_keyword_type = select_step_keyword_type


class WhenStep(StepPrototype):
    steps: Sequence[
        Union["Alternative", "Step", "StarStep", "GivenStep", "WhenStep", "ThenStep", "AndStep", "ButStep", "SubStep"]
    ] = Field(default_factory=list, alias="Steps")
    type: Optional[Keyword] = Field(Keyword.When)
    keyword_type: Optional[KeywordType] = Field(KeywordType.action, alias="Type")
    action: Optional[str] = Field(alias=Keyword.When.value)

    convert_to_step = convert_to_step
    select_step_keyword_type = select_step_keyword_type


class ThenStep(StepPrototype):
    steps: Sequence[
        Union["Alternative", "Step", "StarStep", "GivenStep", "WhenStep", "ThenStep", "AndStep", "ButStep", "SubStep"]
    ] = Field(default_factory=list, alias="Steps")
    type: Optional[Keyword] = Field(Keyword.Then)
    keyword_type: Optional[KeywordType] = Field(KeywordType.outcome, alias="Type")
    action: Optional[str] = Field(alias=Keyword.Then.value)

    convert_to_step = convert_to_step
    select_step_keyword_type = select_step_keyword_type


class AndStep(StepPrototype):
    steps: Sequence[
        Union["Alternative", "Step", "StarStep", "GivenStep", "WhenStep", "ThenStep", "AndStep", "ButStep", "SubStep"]
    ] = Field(default_factory=list, alias="Steps")
    type: Optional[Keyword] = Field(Keyword.And)
    keyword_type: Optional[KeywordType] = Field(KeywordType.conjunction, alias="Type")
    action: Optional[str] = Field(alias=Keyword.And.value)

    convert_to_step = convert_to_step
    select_step_keyword_type = select_step_keyword_type


class ButStep(StepPrototype):
    steps: Sequence[
        Union["Alternative", "Step", "StarStep", "GivenStep", "WhenStep", "ThenStep", "AndStep", "ButStep", "SubStep"]
    ] = Field(default_factory=list, alias="Steps")
    type: Optional[Keyword] = Field(Keyword.But)
    keyword_type: Optional[KeywordType] = Field(KeywordType.conjunction, alias="Type")
    action: Optional[str] = Field(alias=Keyword.But.value)

    convert_to_step = convert_to_step
    select_step_keyword_type = select_step_keyword_type


Join.update_forward_refs()
Step.update_forward_refs()
StarStep.update_forward_refs()
GivenStep.update_forward_refs()
WhenStep.update_forward_refs()
ThenStep.update_forward_refs()
AndStep.update_forward_refs()
ButStep.update_forward_refs()
Alternative.update_forward_refs()

Given = GivenStep.build_action_first_arg
When = WhenStep.build_action_first_arg
Then = ThenStep.build_action_first_arg
And = AndStep.build_action_first_arg
But = ButStep.build_action_first_arg



"""
Compatibility module for pytest
"""
from __future__ import annotations

import sys
from operator import ge
from pathlib import Path
from typing import TYPE_CHECKING, cast

from _pytest.config import Config, PytestPluginManager
from _pytest.config.argparsing import Parser
from _pytest.fixtures import FixtureDef, FixtureLookupError, call_fixture_func
from _pytest.main import Session, wrap_session
from _pytest.mark import Mark, MarkDecorator
from _pytest.python import Metafunc
from _pytest.reports import TestReport
from _pytest.runner import CallInfo
from _pytest.terminal import TerminalReporter
from pytest import Module as PytestModule
from pytest import fail as _pytest_fail

from pytest_bdd.packaging import compare_distribution_version

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias


# region pytest version dependent imports
def is_pytest_version_greater(version: str):
    return compare_distribution_version("pytest", version, ge)


PYTEST6, PYTEST61, PYTEST62, PYTEST7 = map(
    is_pytest_version_greater,
    [
        "6.0",
        "6.1",
        "6.2",
        "7.0",
    ],
)

if PYTEST6:
    from _pytest.config import ExitCode
else:
    ExitCode: TypeAlias = int  # type:ignore[no-redef]

if PYTEST7:
    if TYPE_CHECKING:
        from pytest import Testdir
else:
    import py

    if TYPE_CHECKING:
        from _pytest.pytester import Testdir  # type: ignore[no-redef, attr-defined]

if PYTEST62:
    from pytest import FixtureRequest
else:
    from _pytest.fixtures import FixtureRequest
# endregion

if TYPE_CHECKING:  # pragma: no cover
    from _pytest.nodes import Item as BaseItem
    from _pytest.pytester import RunResult

    class Item(BaseItem):
        _request: FixtureRequest

else:
    from _pytest.nodes import Item


class Module(PytestModule):
    @classmethod
    def build(cls, parent, file_path):
        if hasattr(cls, "from_parent"):
            collector = cls.from_parent(
                parent, **(dict(path=Path(file_path)) if PYTEST7 else dict(fspath=py.path.local(file_path)))
            )
        else:
            collector = cls(parent=parent, fspath=py.path.local(file_path))
        return collector

    def get_path(self):
        return getattr(self, "path", Path(self.fspath))


if PYTEST6:

    def assert_outcomes(
        result: RunResult,
        passed: int = 0,
        skipped: int = 0,
        failed: int = 0,
        errors: int = 0,
        xpassed: int = 0,
        xfailed: int = 0,
    ) -> None:
        """Compatibility function for result.assert_outcomes"""
        result.assert_outcomes(
            errors=errors, passed=passed, skipped=skipped, failed=failed, xpassed=xpassed, xfailed=xfailed
        )

else:

    def assert_outcomes(
        result: RunResult,
        passed: int = 0,
        skipped: int = 0,
        failed: int = 0,
        errors: int = 0,
        xpassed: int = 0,
        xfailed: int = 0,
    ) -> None:
        """Compatibility function for result.assert_outcomes"""
        result.assert_outcomes(  # type: ignore[call-arg]
            #  Pytest < 6 uses the singular form
            error=errors,
            passed=passed,
            skipped=skipped,
            failed=failed,
            xpassed=xpassed,
            xfailed=xfailed,
        )


def get_config_root_path(config: Config) -> Path:
    return Path(getattr(cast(Config, config), "rootpath" if PYTEST61 else "rootdir"))


def fail(reason, pytrace=True):
    __tracebackhide__ = True
    if PYTEST7:
        return _pytest_fail(reason, pytrace=pytrace)
    else:
        return _pytest_fail(msg=reason, pytrace=pytrace)


__all__ = [
    "assert_outcomes",
    "Item",
    "CallInfo",
    "call_fixture_func",
    "Config",
    "ExitCode",
    "fail",
    "FixtureDef",
    "FixtureLookupError",
    "FixtureRequest",
    "get_config_root_path",
    "Mark",
    "MarkDecorator",
    "Metafunc",
    "Module",
    "Parser",
    "PytestPluginManager",
    "PYTEST6",
    "PYTEST7",
    "RunResult",
    "Session",
    "TerminalReporter",
    "Testdir",
    "TestReport",
    "TypeAlias",
    "wrap_session",
]



import pytest
from _pytest.outcomes import Failed
from attr import attrib, attrs
from pytest import raises

from pytest_bdd.utils import deepattrgetter, doesnt_raise, setdefaultattr


def test_get_attribute():
    @attrs
    class Foo:
        foo = attrib()

    item = Foo(foo=1)

    assert deepattrgetter("foo")(item) == (1,)


def test_get_multi_attribute():
    @attrs
    class Foo:
        foo = attrib()
        boo = attrib()

    item = Foo(foo=1, boo=2)

    assert deepattrgetter("boo", "foo")(item) == (2, 1)


def test_get_default_for_attribute():
    class Foo:
        ...

    item = Foo()

    assert deepattrgetter("foo", default=1)(item) == (1,)


def test_get_default_for_multi_attribute():
    class Foo:
        ...

    item = Foo()

    assert deepattrgetter("boo", "foo", default=1)(item) == (1, 1)


def test_raise_on_missing_attribute():
    class Foo:
        ...

    item = Foo()

    with raises(AttributeError):
        deepattrgetter("foo")(item)


def test_raise_on_missing_multi_attribute():
    class Foo:
        ...

    item = Foo()

    with raises(AttributeError):
        deepattrgetter("foo", "boo")(item)


def test_raise_on_missing_nested_attribute():
    class Foo:
        ...

    item = Foo()

    with raises(AttributeError):
        deepattrgetter("foo.bar")(item)


def test_raise_on_missing_multi_nested_attribute():
    class Foo:
        ...

    item = Foo()

    with raises(AttributeError):
        deepattrgetter("foo.bar", "boo.car")(item)


def test_get_nested_attribute():
    @attrs
    class Foo:
        foo = attrib()

    @attrs
    class Bar:
        bar = attrib()

    item = Foo(foo=Bar(bar=1))

    assert deepattrgetter("foo.bar")(item) == (1,)


def test_get_nested_multi_attribute():
    @attrs
    class Foo:
        foo = attrib()
        boo = attrib()

    @attrs
    class Bar:
        bar = attrib()

    @attrs
    class Boo:
        boo = attrib()

    item = Foo(foo=Bar(bar=1), boo=Boo(2))

    assert deepattrgetter("boo.boo", "foo.bar")(item) == (2, 1)


def test_skip_missing_attributes():
    @attrs
    class Foo:
        foo = attrib()
        boo = attrib()

    item = Foo(foo="bar", boo="baz")

    assert deepattrgetter("foo", "boo", "missing", skip_missing=True)(item) == ("bar", "baz")


def test_skip_missing_attributes_nested():
    @attrs
    class Foo:
        foo = attrib()
        boo = attrib()

    @attrs
    class Bar:
        bar = attrib()

    @attrs
    class Boo:
        boo = attrib()

    item = Foo(foo=Bar(bar=1), boo=Boo(2))

    assert deepattrgetter("boo.boo", "foo.bar", "missing.missing", skip_missing=True)(item) == (2, 1)


def test_skip_missing_and_default_attributes():
    @attrs
    class Foo:
        foo = attrib()
        boo = attrib()

    item = Foo(foo="bar", boo="baz")

    with raises(ValueError):
        deepattrgetter("foo", "boo", "missing", skip_missing=True, default=object())(item)


def test_setdefaultattr_set_nonexisting_attr_value():
    class Dumb:
        ...

    dumb = Dumb()
    setdefaultattr(dumb, "foo", value=10)

    assert dumb.foo == 10


def test_setdefaultattr_set_nonexisting_attr_value_factory():
    class Dumb:
        ...

    dumb = Dumb()
    setdefaultattr(dumb, "foo", value_factory=lambda: 10)

    assert dumb.foo == 10


def test_setdefaultattr_not_set_existing_attr_value():
    class Dumb:
        ...

    dumb = Dumb()
    dumb.foo = 20
    setdefaultattr(dumb, "foo", value=10)

    assert dumb.foo == 20


def test_setdefaultattr_not_set_existing_attr_value_factory():
    class Dumb:
        ...

    dumb = Dumb()
    dumb.foo = 20
    setdefaultattr(dumb, "foo", value_factory=lambda: 10)

    assert dumb.foo == 20


def test_setdefaultattr_for_both_factory_and_value():
    class Dumb:
        ...

    with pytest.raises(ValueError):
        setdefaultattr(Dumb(), "a", value=10, value_factory=lambda: 20)


def test_doesnt_raise_fails_test():
    with raises(Failed):
        with doesnt_raise(RuntimeError):
            raise RuntimeError


def test_doesnt_raise_suppress_if_not_match():
    try:
        with doesnt_raise(RuntimeError, match="cool"):
            raise RuntimeError("nice")
    except Exception as e:  # pragma: no cover
        raise AssertionError from e


def test_doesnt_raise_not_suppress_if_not_match_explicitly():
    with raises(RuntimeError, match="nice"):
        with doesnt_raise(RuntimeError, match="cool", suppress_not_matched=False):
            raise RuntimeError("nice")


def test_doesnt_raise_passes_original_exception_if_not_suppressed():
    with raises(ValueError):
        with doesnt_raise(RuntimeError, suppress_not_matched=False):
            raise ValueError("nice")



from configparser import ConfigParser
from importlib.machinery import ModuleSpec
from importlib.util import module_from_spec
from pathlib import Path
from typing import Optional, Tuple, cast
from uuid import uuid4

from pytest_bdd.compatibility.pytest import Module as PytestModule
from pytest_bdd.model.messages import Message
from pytest_bdd.scenario import scenarios
from pytest_bdd.steps import StepHandler
from pytest_bdd.utils import convert_str_to_python_name
from pytest_bdd.webloc import read as webloc_read


class Module(PytestModule):
    def collect(self):
        StepHandler.Registry.inject_registry_fixture_and_register_steps(self.obj)
        try:
            registry = self.obj.step_registry.__registry__.registry
        except AttributeError:
            registry = set()
        config = self.session.config
        if config.option.messages_ndjson_path is not None:
            for step_definition in registry:
                config.hook.pytest_bdd_message(
                    config=config, message=Message(stepDefinition=step_definition.as_message(config=config))
                )
        return super().collect()


class FeatureFileModule(Module):
    def _getobj(self):
        path: Path = self.get_path()
        if ".url" == path.suffixes[-1]:
            feature_pathlike, base_dir = self.get_feature_pathlike_from_url_file(path)
        elif ".desktop" == path.suffixes[-1]:
            feature_pathlike, base_dir = self.get_feature_pathlike_from_desktop_file(path), None
        elif ".webloc" == path.suffixes[-1]:
            feature_pathlike, base_dir = self.get_feature_pathlike_from_weblock_file(path), None
        else:
            feature_pathlike, base_dir = path, None
        return self._build_test_module(feature_pathlike, base_dir)

    def _build_test_module(self, path: Optional[Path], base_dir: Optional[Path]):
        module_name = convert_str_to_python_name(f"{path}_{uuid4()}")

        module_spec = ModuleSpec(module_name, None)
        module = module_from_spec(module_spec)

        module.test_scenarios = scenarios(  # type:ignore[attr-defined]
            *((path,) if path is not None else []),
            filter_=None,
            return_test_decorator=False,
            parser_type=getattr(self, "parser_type", None),
            features_base_dir=base_dir,
        )

        return module

    @staticmethod
    def get_feature_pathlike_from_url_file(path: Path) -> Tuple[str, Optional[str]]:
        config_parser = ConfigParser()
        config_parser.read(path)

        config_data = config_parser["InternetShortcut"]
        working_dir = config_data.get("WorkingDirectory", None)
        url = config_data.get("URL", None)
        return url, working_dir

    @staticmethod
    def get_feature_pathlike_from_desktop_file(path: Path) -> Optional[str]:
        config_parser = ConfigParser()
        config_parser.read(path)

        config_data = config_parser["Desktop Entry"]
        return config_data["URL"] if config_data["Type"] == "Link" else None

    @staticmethod
    def get_feature_pathlike_from_weblock_file(path: Path) -> str:
        return cast(str, webloc_read(str(path)))



from pytest_bdd.model.gherkin_document import Feature
from pytest_bdd.model.messages import Pickle, PickleStep, Step, Tag
from pytest_bdd.model.messages import Type as StepType

__all__ = ["Feature", "Pickle", "Step", "StepType", "Tag"]



import sys

if sys.version_info >= (3, 10):
    from typing import Literal, Protocol, TypeAlias, runtime_checkable
else:
    from typing import runtime_checkable

    from typing_extensions import Literal, Protocol, TypeAlias

__all__ = [
    "Literal",
    "Protocol",
    "runtime_checkable",
    "TypeAlias",
]



import linecache
from functools import partial
from itertools import filterfalse
from operator import contains, methodcaller
from os.path import relpath
from pathlib import Path
from typing import Callable, List, Sequence, Set, Union, cast

from attr import attrib, attrs
from attr._make import Factory
from gherkin.ast_builder import AstBuilder
from gherkin.errors import CompositeParserException
from gherkin.parser import Parser as CucumberIOBaseParser  # type: ignore[import]
from gherkin.pickles.compiler import Compiler as PicklesCompiler

from pytest_bdd.compatibility.parser import ParserProtocol
from pytest_bdd.compatibility.pytest import Config
from pytest_bdd.compatibility.struct_bdd import STRUCT_BDD_INSTALLED
from pytest_bdd.exceptions import FeatureError
from pytest_bdd.model import Feature
from pytest_bdd.model.messages import MediaType, Message, Source
from pytest_bdd.utils import IdGenerator, PytestBDDIdGeneratorHandler

if STRUCT_BDD_INSTALLED:  # pragma: no cover
    from pytest_bdd.struct_bdd.parser import StructBDDParser

    assert StructBDDParser  # type: ignore[truthy-function]


@attrs
class GlobMixin:
    glob: Callable[..., Sequence[Union[str, Path]]] = attrib(default=methodcaller("glob", "*.feature"), kw_only=True)


class ASTBuilderMixin:
    def build_feature(self, gherkin_document_raw_dict, filename: str, id_generator) -> Feature:
        gherkin_document = Feature.load_gherkin_document(gherkin_document_raw_dict)

        pickles_data = PicklesCompiler(id_generator=id_generator).compile(gherkin_document_raw_dict)
        pickles = Feature.load_pickles(pickles_data)

        feature = Feature(  # type: ignore[call-arg]
            gherkin_document=gherkin_document,
            uri=gherkin_document.uri,
            pickles=pickles,
            filename=filename,
        )

        return feature


@attrs
class GherkinParser(CucumberIOBaseParser, ASTBuilderMixin, GlobMixin, ParserProtocol):
    id_generator = attrib(default=Factory(IdGenerator))

    def __attrs_post_init__(self):
        CucumberIOBaseParser.__init__(self, ast_builder=AstBuilder(id_generator=self.id_generator))

    @classmethod
    def parse(
        cls, config: Union[Config, PytestBDDIdGeneratorHandler], path: Path, uri: str, *args, **kwargs
    ) -> Feature:
        parser = cls(id_generator=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator)
        encoding = kwargs.pop("encoding", "utf-8")
        with path.open(mode="r", encoding=encoding) as feature_file:
            feature_file_data = feature_file.read()

        if ".md" in path.suffixes:
            media_type = MediaType.text_x_cucumber_gherkin_markdown
        else:
            media_type = MediaType.text_x_cucumber_gherkin_plain

        cast(Config, config).hook.pytest_bdd_message(
            config=config, message=Message(source=Source(uri=uri, data=feature_file_data, media_type=media_type))
        )

        try:
            gherkin_document_raw_dict = CucumberIOBaseParser.parse(
                parser, token_scanner_or_str=feature_file_data, *args, **kwargs
            )
        except CompositeParserException as e:
            raise FeatureError(
                e.args[0],
                e.errors[0].location["line"],
                linecache.getline(str(path), e.errors[0].location["line"]).rstrip("\n"),
                uri,
            ) from e

        gherkin_document_raw_dict["uri"] = uri

        feature = parser.build_feature(
            gherkin_document_raw_dict,
            filename=str(path.as_posix()),
            id_generator=getattr(config, "pytest_bdd_id_generator", IdGenerator()),
        )

        cast(Config, config).hook.pytest_bdd_message(
            config=config, message=Message(gherkin_document=feature.gherkin_document)
        )
        for pickle in feature.pickles:
            cast(Config, config).hook.pytest_bdd_message(config=config, message=Message(pickle=pickle))

        return feature

    def get_from_paths(self, config: Config, paths: Sequence[Path], **kwargs) -> Sequence[Feature]:
        """Get features for given paths."""
        seen_names: Set[Path] = set()
        features: List[Feature] = []
        features_base_dir = kwargs.pop("features_base_dir", Path.cwd())
        if not features_base_dir.is_absolute():
            features_base_dir = Path.cwd() / features_base_dir

        for rel_path in map(Path, paths):
            path = rel_path if rel_path.is_absolute() else Path(features_base_dir) / rel_path

            file_paths = list(map(Path, self.glob(path))) if path.is_dir() else [Path(path)]

            features.extend(
                map(
                    lambda path: self.parse(
                        config, path, "file:" + relpath(str(path), str(features_base_dir)), **kwargs
                    ),
                    filterfalse(partial(contains, seen_names), file_paths),
                )
            )

            for file_path in file_paths:
                if file_path not in seen_names:
                    seen_names.add(path)

        return sorted(features, key=lambda feature: feature.name or feature.filename)



from pathlib import Path
from typing import TYPE_CHECKING, Callable, Sequence, Union

from pytest_bdd.compatibility.pytest import Config
from pytest_bdd.compatibility.typing import Protocol, runtime_checkable
from pytest_bdd.utils import PytestBDDIdGeneratorHandler

if TYPE_CHECKING:  # pragma: no cover
    from pytest_bdd.model import Feature


@runtime_checkable
class ParserProtocol(Protocol):
    # Defines which files would be parsed
    glob: Callable[[Path], Sequence[Union[str, Path]]]

    def __init__(self, *args, id_generator=None, **kwargs):  # pragma: no cover
        ...

    def parse(
        self, config: Union[Config, PytestBDDIdGeneratorHandler], path: Path, uri: str, *args, **kwargs
    ) -> "Feature":  # pragma: no cover
        ...



"""Reporting functionality.

Collection of the scenario execution statuses, timing and other information
that enriches the pytest test reporting.
"""
import time
from typing import Any, Callable, Dict, List, Sequence

import pytest
from attr import Factory, attrib, attrs

from pytest_bdd.compatibility.pytest import CallInfo, FixtureRequest, Item
from pytest_bdd.model import Feature
from pytest_bdd.model.messages import Pickle, PickleStep


class StepReport:
    """StepHandler execution report."""

    failed = False
    stopped = None

    def __init__(self, step: PickleStep) -> None:
        """StepHandler report constructor.

        :param StepHandler step: StepHandler.
        """
        self.step = step
        self.started = time.perf_counter()

    def serialize(self, feature: Feature) -> Dict[str, Any]:
        """Serialize the step execution report.

        :return: Serialized step execution report.
        :rtype: dict
        """
        return {
            "name": self.step.text,
            "type": feature._get_step_prefix(self.step),
            "keyword": feature._get_step_keyword(self.step),
            "line_number": feature._get_step_line_number(self.step),
            "failed": self.failed,
            "duration": self.duration,
        }

    def finalize(self, failed: bool) -> None:
        """Stop collecting information and finalize the report.

        :param bool failed: Whether the step execution is failed.
        """
        self.stopped = time.perf_counter()
        self.failed = failed

    @property
    def duration(self) -> float:
        """StepHandler execution duration.

        :return: StepHandler execution duration.
        :rtype: float
        """
        if self.stopped is None:
            return 0

        return self.stopped - self.started


@attrs
class ScenarioReport:
    """Pickle execution report."""

    feature: Feature = attrib()
    scenario: Pickle = attrib()
    step_reports: List[StepReport] = attrib(default=Factory(list))

    @property
    def current_step_report(self) -> StepReport:
        """Get current step report.

        :return: Last or current step report.
        :rtype: pytest_bdd.reporting.StepReport
        """
        return self.step_reports[-1]

    def add_step_report(self, step_report: StepReport) -> None:
        """Add new step report.

        :param step_report: New current step report.
        :type step_report: pytest_bdd.reporting.StepReport
        """
        self.step_reports.append(step_report)

    def serialize(self) -> Dict[str, Any]:
        """Serialize scenario execution report in order to transfer reporting from nodes in the distributed mode.

        :return: Serialized report.
        :rtype: dict
        """
        pickle = self.scenario
        feature: Feature = self.feature

        return {
            "steps": [step_report.serialize(self.feature) for step_report in self.step_reports],
            "name": pickle.name,
            "line_number": feature._get_pickle_line_number(pickle),
            "tags": sorted(set(feature._get_pickle_tag_names(pickle)).difference(feature.tag_names)),
            "feature": {
                "name": feature.name,
                "filename": feature.filename,
                "rel_filename": feature.rel_filename,
                "line_number": feature.line_number,
                "description": feature.description,
                "tags": feature.tag_names,
            },
        }

    def fail(self) -> None:
        """Stop collecting information and finalize the report as failed."""
        self.current_step_report.finalize(failed=True)
        remaining_steps = self.scenario.steps[len(self.step_reports) :]

        # Fail the rest of the steps and make reports.
        for step in remaining_steps:
            report = StepReport(step=step)
            report.finalize(failed=True)
            self.add_step_report(report)


class ScenarioReporterPlugin:
    def __init__(self):
        self.current_report = None

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(self, item: Item, call: CallInfo):
        outcome = yield
        if call.when != "setup":
            rep = outcome.get_result()
            """Store item in the report object."""
            scenario_report: ScenarioReport = self.current_report

            if scenario_report is not None:
                rep.scenario = scenario_report.serialize()
                rep.item = {"name": item.name}

    @pytest.hookimpl(tryfirst=True)
    def pytest_bdd_before_scenario(self, request: FixtureRequest, feature: Feature, scenario: Pickle) -> None:
        """Create scenario report for the item."""
        self.current_report = ScenarioReport(feature=feature, scenario=scenario)  # type: ignore[call-arg]

    @pytest.hookimpl(tryfirst=True)
    def pytest_bdd_step_error(
        self,
        request: FixtureRequest,
        feature: Feature,
        scenario: Pickle,
        step: PickleStep,
        step_func: Callable,
        step_func_args: dict,
        exception: Exception,
    ) -> None:
        """Finalize the step report as failed."""
        self.current_report.fail()

    @pytest.hookimpl(tryfirst=True)
    def pytest_bdd_before_step(
        self, request: FixtureRequest, feature: Feature, scenario: Pickle, step: PickleStep, step_func: Callable
    ) -> None:
        """Store step start time."""
        self.current_report.add_step_report(StepReport(step=step))

    @pytest.hookimpl(tryfirst=True)
    def pytest_bdd_after_step(
        self,
        request: FixtureRequest,
        feature: Feature,
        scenario: Pickle,
        step: PickleStep,
        step_func: Callable,
        step_func_args: dict,
    ) -> None:
        """Finalize the step report as successful."""
        self.current_report.current_step_report.finalize(failed=False)



from pathlib import Path
from time import time_ns
from typing import Union, cast

from _pytest.fixtures import FixtureRequest
from _pytest.main import Session
from attr import attrib, attrs
from pytest import ExitCode

from pytest_bdd.compatibility.pytest import Config, Parser
from pytest_bdd.model.messages import (
    Duration,
    Message,
    Status,
    TestCase,
    TestCaseFinished,
    TestRunFinished,
    TestRunStarted,
    TestStep,
    TestStepFinished,
    TestStepResult,
    TestStepStarted,
    Timestamp,
)
from pytest_bdd.steps import StepHandler
from pytest_bdd.utils import PytestBDDIdGeneratorHandler, deepattrgetter


@attrs(eq=False)
class MessagePlugin:
    config: Config = attrib()
    current_test_case = attrib(default=None)
    current_test_case_step_to_definition_mapping = attrib(default=None)

    def get_timestamp(self):
        timestamp = time_ns()
        test_run_started_seconds = timestamp // 10**9
        test_run_started_nanos = timestamp - test_run_started_seconds * 10**9
        return Timestamp(seconds=test_run_started_seconds, nanos=test_run_started_nanos)

    @staticmethod
    def add_options(parser: Parser) -> None:
        """Add pytest-bdd options."""
        group = parser.getgroup("bdd", "Cucumber NDJSON")
        group.addoption(
            "--messagesndjson",
            "--messages-ndjson",
            action="store",
            dest="messages_ndjson_path",
            metavar="path",
            default=None,
            help="messages ndjson report file at given path.",
        )

    def pytest_bdd_message(self, config: Config, message: Message):
        if config.option.messages_ndjson_path is None:
            return
        with Path(self.config.option.messages_ndjson_path).open(mode="a+") as f:
            f.write(message.json(exclude_none=True))
            f.write("\n")

    def pytest_runtestloop(self, session: Session):
        config = session.config
        if config.option.messages_ndjson_path is None:
            return
        hook_handler = config.hook

        # TODO check messaging of step definitions; seems outdated

        # Message plugins step definitions (test module step definitions are messaged separately)
        for plugin in session.config.pluginmanager._plugin2hookcallers.keys():
            registry = deepattrgetter("step_registry.__registry__.registry", default=set())(plugin)[0]
            for step_definition in registry:
                hook_handler.pytest_bdd_message(
                    config=config, message=Message(step_definition=step_definition.as_message(config=config))
                )

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(test_run_started=TestRunStarted(timestamp=self.get_timestamp())),
        )

    def pytest_sessionfinish(self, session, exitstatus):
        config = session.config
        if config.option.messages_ndjson_path is None:
            return
        hook_handler = config.hook

        is_testrun_success = (isinstance(exitstatus, int) and exitstatus == 0) or exitstatus is ExitCode.OK
        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                test_run_finished=TestRunFinished(
                    timestamp=self.get_timestamp(),
                    success=is_testrun_success,
                )
            ),
        )

    def pytest_bdd_before_scenario(self, request: FixtureRequest, feature, scenario):
        config: Union[
            Config, PytestBDDIdGeneratorHandler
        ] = request.config  # https://github.com/python/typing/issues/213
        if cast(Config, config).option.messages_ndjson_path is None:
            return
        hook_handler = cast(Config, config).hook

        test_steps = []
        previous_step = None
        self.current_test_case_step_id_to_step_mapping = {}
        for step in scenario.steps:
            try:
                step_definition = hook_handler.pytest_bdd_match_step_definition_to_step(
                    request=request,
                    feature=feature,
                    scenario=scenario,
                    step=step,
                    previous_step=previous_step,
                )
            except StepHandler.Matcher.MatchNotFoundError as e:
                pass
            else:
                test_step = TestStep(
                    id=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator.get_next_id(),
                    pickle_step_id=step.id,
                    step_definition_ids=[step_definition.id]
                    # TODO Check step_match_arguments_lists
                )
                test_steps.append(test_step)
                self.current_test_case_step_id_to_step_mapping[id(step)] = test_step
            finally:
                previous_step = step

        self.current_test_case = TestCase(
            id=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator.get_next_id(),
            pickle_id=scenario.id,
            test_steps=test_steps,
        )

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(test_case=self.current_test_case),
        )

    def pytest_bdd_after_scenario(self, request, feature, scenario):
        config = request.config
        if config.option.messages_ndjson_path is None:
            return

        hook_handler = config.hook

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                test_case_finished=TestCaseFinished(
                    test_case_started_id=self.current_test_case.id,
                    timestamp=self.get_timestamp(),
                    # TODO check usage
                    will_be_retried=False,
                )
            ),
        )
        self.current_test_case = None

    def pytest_bdd_before_step(self, request, feature, scenario, step, step_func):
        config = request.config
        if config.option.messages_ndjson_path is None:
            return
        hook_handler = config.hook

        # TODO check behaviour if missing
        step_definition = self.current_test_case_step_id_to_step_mapping[id(step)]

        self.current_test_case_step_start_timestamp = self.get_timestamp()

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                test_step_started=TestStepStarted(
                    test_case_started_id=self.current_test_case.id,
                    timestamp=self.current_test_case_step_start_timestamp,
                    test_step_id=step_definition.id,
                )
            ),
        )

    def pytest_bdd_after_step(self, request, feature, scenario, step, step_func):
        config = request.config
        if config.option.messages_ndjson_path is None:
            return
        hook_handler = config.hook

        # TODO check behaviour if missing
        step_definition = self.current_test_case_step_id_to_step_mapping[id(step)]
        self.current_test_case_step_finish_timestamp = self.get_timestamp()

        current_test_case_step_duration_total_nanos = (
            self.current_test_case_step_finish_timestamp.seconds * 10**9
            + self.current_test_case_step_finish_timestamp.nanos
        ) - (
            self.current_test_case_step_start_timestamp.seconds * 10**9
            + self.current_test_case_step_start_timestamp.nanos
        )
        current_test_case_step_duration_seconds = current_test_case_step_duration_total_nanos // 10**9
        current_test_case_step_duration_nanos = (
            current_test_case_step_duration_total_nanos - current_test_case_step_duration_seconds * 10**9
        )
        current_test_case_step_duration = Duration(
            seconds=current_test_case_step_duration_seconds, nanos=current_test_case_step_duration_nanos
        )

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                test_step_finished=TestStepFinished(
                    test_case_started_id=self.current_test_case.id,
                    timestamp=self.current_test_case_step_finish_timestamp,
                    test_step_id=step_definition.id,
                    test_step_result=TestStepResult(duration=current_test_case_step_duration, status=Status.passed),
                )
            ),
        )

    def pytest_bdd_step_error(
        self, request, feature, scenario, step, step_func, step_func_args, exception, step_definition
    ):
        config = request.config
        if config.option.messages_ndjson_path is None:
            return
        hook_handler = config.hook

        # TODO check behaviour if missing
        step_definition = self.current_test_case_step_id_to_step_mapping[id(step)]
        self.current_test_case_step_finish_timestamp = self.get_timestamp()

        current_test_case_step_duration_total_nanos = (
            self.current_test_case_step_finish_timestamp.seconds * 10**9
            + self.current_test_case_step_finish_timestamp.nanos
        ) - (
            self.current_test_case_step_start_timestamp.seconds * 10**9
            + self.current_test_case_step_start_timestamp.nanos
        )
        current_test_case_step_duration_seconds = current_test_case_step_duration_total_nanos // 10**9
        current_test_case_step_duration_nanos = (
            current_test_case_step_duration_total_nanos - current_test_case_step_duration_seconds * 10**9
        )
        current_test_case_step_duration = Duration(
            seconds=current_test_case_step_duration_seconds, nanos=current_test_case_step_duration_nanos
        )

        hook_handler.pytest_bdd_message(
            config=config,
            message=Message(
                test_step_finished=TestStepFinished(
                    test_case_started_id=self.current_test_case.id,
                    timestamp=self.current_test_case_step_finish_timestamp,
                    test_step_id=step_definition.id,
                    test_step_result=TestStepResult(duration=current_test_case_step_duration, status=Status.failed),
                )
            ),
        )



"""Various utility functions."""
import base64
import pickle
import re
import sys
from collections import defaultdict
from contextlib import contextmanager, nullcontext, suppress
from enum import Enum
from functools import reduce
from inspect import getframeinfo, signature
from itertools import tee
from operator import attrgetter, getitem, itemgetter
from sys import _getframe
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Dict,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Type,
    Union,
    cast,
)

from pytest import raises

from pytest_bdd.compatibility.pytest import FixtureDef, fail
from pytest_bdd.compatibility.typing import Literal, Protocol, runtime_checkable
from pytest_bdd.const import ALPHA_REGEX, PYTHON_REPLACE_REGEX

if TYPE_CHECKING:  # pragma: no cover
    from pytest_bdd.compatibility.pytest import RunResult


@runtime_checkable
class PytestBDDIdGeneratorHandler(Protocol):
    pytest_bdd_id_generator: Union["IdGenerator", Any]


def get_args(func: Callable) -> Sequence[str]:
    """Get a list of argument names for a function.

    :param func: The function to inspect.

    :return: A list of argument names.
    :rtype: list
    """
    params = signature(func).parameters.values()
    return [param.name for param in params if param.kind == param.POSITIONAL_OR_KEYWORD]


def get_caller_module_locals(stacklevel: int = 1) -> Dict[str, Any]:
    """Get the caller module locals dictionary.

    We use sys._getframe instead of inspect.stack(0) because the latter is way slower, since it iterates over
    all the frames in the stack.
    """
    return _getframe(stacklevel).f_locals


def get_caller_module_path(stacklevel: int = 1) -> str:
    """Get the caller module path.

    We use sys._getframe instead of inspect.stack(0) because the latter is way slower, since it iterates over
    all the frames in the stack.
    """
    frame = _getframe(stacklevel)
    return getframeinfo(frame, context=0).filename


def convert_str_to_python_name(s: Any) -> str:
    s1: str = str(s)
    s2 = re.sub(r"[^.a-zA-Z0-9]", "_", s1)
    s3 = re.sub(r"_+", "_", s2)
    s4 = s3.strip("_")
    if re.match(r"\d.*", s4):
        result_s = f"_{s4}"
    else:
        result_s = s4
    return result_s


_DUMP_START = "_pytest_bdd_>>>"
_DUMP_END = "<<<_pytest_bdd_"


def dump_obj(*objects: Any) -> None:
    """Dump objects to stdout so that they can be inspected by the test suite."""
    for obj in objects:
        dump = pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
        encoded = base64.b64encode(dump).decode("ascii")
        print(f"{_DUMP_START}{encoded}{_DUMP_END}")


def collect_dumped_objects(result: "RunResult"):
    """Parse all the objects dumped with `dump_object` from the result.

    Note: You must run the result with output to stdout enabled.
    For example, using ``testdir.runpytest("-s")``.
    """
    stdout = result.stdout.str()  # pytest < 6.2, otherwise we could just do str(result.stdout)
    payloads = re.findall(rf"{_DUMP_START}(.*?){_DUMP_END}", stdout)
    return [pickle.loads(base64.b64decode(payload)) for payload in payloads]


class DefaultMapping(defaultdict):
    Skip = object()

    def __init__(self, *args, default_factory=None, warm_up_keys=(), **kwargs):
        super().__init__(default_factory, *args, **kwargs)
        self.warm_up(*warm_up_keys)

    def __missing__(self, key):
        if ... in self.keys():
            intercessor = self[...]
            if intercessor is self.Skip:
                raise KeyError(key)
            elif isinstance(intercessor, Callable):
                value = intercessor(key)
            elif intercessor is ...:
                value = key
            else:
                value = intercessor
            self[key] = value
            return value
        else:
            return super().__missing__(key)

    def warm_up(self, *items):
        for item in items:
            with suppress(KeyError):
                getitem(self, item)

    @classmethod
    def instantiate_from_collection_or_bool(
        cls, bool_or_items: Union[Collection[str], Dict[str, Any], Any] = True, *, warm_up_keys=()
    ):
        if isinstance(bool_or_items, Collection):
            if not isinstance(bool_or_items, Mapping):
                bool_or_items = zip(*tee(iter(bool_or_items)))
        else:
            bool_or_items = cast(dict, {...: ...} if bool_or_items else {...: DefaultMapping.Skip})
        return cls(bool_or_items, warm_up_keys=warm_up_keys)


def inject_fixture(request, arg, value):
    """Inject fixture into pytest fixture request.

    :param request: pytest fixture request
    :param arg: argument name
    :param value: argument value
    """

    fd = FixtureDef(
        fixturemanager=request._fixturemanager,
        baseid=None,
        argname=arg,
        func=lambda: value,
        scope="function",
        params=None,
    )
    fd.cached_result = (value, 0, None)

    old_fd = request._fixture_defs.get(arg)
    add_fixturename = arg not in request.fixturenames

    def fin():
        request._fixturemanager._arg2fixturedefs[arg].remove(fd)
        request._fixture_defs[arg] = old_fd

        if add_fixturename:
            request._pyfuncitem._fixtureinfo.names_closure.remove(arg)

    request.addfinalizer(fin)

    # inject fixture definition
    request._fixturemanager._arg2fixturedefs.setdefault(arg, []).insert(0, fd)
    # inject fixture value in request cache
    request._fixture_defs[arg] = fd
    if add_fixturename:
        request._pyfuncitem._fixtureinfo.names_closure.append(arg)


def _itemgetter(*items):
    def func(obj):
        if len(items) == 0:
            return []
        elif len(items) == 1:
            return [obj[items[0]]]
        else:
            return itemgetter(*items)(obj)

    return func


class _NoneException(Exception):
    ...


class Empty(Enum):
    empty = None


def getitemdefault(
    obj, index, default=Empty.empty, default_factory: Optional[Callable] = None, treat_as_empty=Empty.empty
):
    if default is not Empty.empty:
        if default_factory is not None:
            raise ValueError("Both 'default' and 'default_factory' were specified")
        else:
            default_factory = lambda: default
    try:
        item = getitem(obj, index)
    except KeyError:
        if default_factory is None:
            raise
        else:
            item = default_factory()
    if item is not treat_as_empty:
        return item
    else:
        raise KeyError(f"{index}")


def deepattrgetter(*attrs, **kwargs):
    empty = object()
    default = kwargs.pop("default", empty)
    default_exception_type = AttributeError if default is not empty else _NoneException
    skip_missing = kwargs.pop("skip_missing", False)
    skip_missing_context = suppress(AttributeError) if skip_missing else nullcontext()
    if default is not empty and skip_missing:
        raise ValueError('Both "default" and "skip_missing" are specified')

    def fn(obj):
        def _():
            for attr in attrs:
                try:
                    with skip_missing_context:
                        yield attrgetter(attr)(obj)
                except default_exception_type:
                    yield default

        return tuple(_())

    return fn


def setdefaultattr(
    obj, key, value: Union[Literal[Empty.empty], Any] = Empty.empty, value_factory: Optional[Callable] = None
):
    if value is not Empty.empty and value_factory is not None:
        raise ValueError("Both 'value' and 'value_factory' were specified")
    with suppress(AttributeError):
        return getattr(obj, key)
    if value_factory is not None:
        value = value_factory()
    setattr(obj, key, value)
    return value


def compose(*funcs):
    return reduce(lambda f, g: lambda *args, **kwargs: f(g(*args, **kwargs)), funcs)


def make_python_name(string: str) -> str:
    """Make python attribute name out of a given string."""
    string = re.sub(PYTHON_REPLACE_REGEX, "", string.replace(" ", "_"))
    return re.sub(ALPHA_REGEX, "", string).lower()


@runtime_checkable
class StringableProtocol(Protocol):
    def __str__(self) -> str:
        ...  # pragma: no cover


def stringify(value: Union[StringableProtocol, str, bytes]) -> str:
    return str(value, **({"encoding": "utf-8"} if isinstance(value, bytes) else {}))


class IdGenerator:
    def __init__(self):
        self._id_counter = 0

    def __next__(self):
        try:
            return str(self._id_counter)
        finally:
            self._id_counter += 1

    get_next_id = __next__


@contextmanager
def doesnt_raise(
    expected_exception: Union[Type[Exception], Sequence[Type[Exception]]],
    *,
    match: Optional[Union[str, Pattern[str]]] = None,
    suppress_not_matched=True,
):
    """

    :param expected_exception: Expected exception/s which don't have to be raised; If it raised - test fails
    :param match: Message which will be count as failing test. If message is not matched - function passes
    :param suppress_not_matched: If specified - all non-matched exceptions will be suppressed
    :return:
    """

    try:
        yield
    except expected_exception:  # type:ignore[misc]
        ex_type, ex_value, ex_traceback = sys.exc_info()
        is_matched = True
        if match is not None:
            is_matched = bool(re.search(match, f"{ex_value}"))
        if is_matched:
            fail(f"{ex_value}")
        elif not suppress_not_matched:
            raise



"""Scenario implementation.

The pytest will collect the test case and the steps will be executed
line by line.

Example:

test_publish_article = scenario(
    feature_name="publish_article.feature",
    scenario_name="Publishing the article",
)
"""
import asyncio
import collections
import os
import ssl
from contextlib import suppress
from enum import Enum
from functools import partial, reduce
from operator import methodcaller, truediv
from os.path import commonpath
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Callable, Iterable, Optional, Tuple, Type, Union, cast
from urllib.parse import urljoin

import aiohttp
import certifi
from attr import Factory, attrib, attrs
from pytest import mark

from pytest_bdd.compatibility.parser import ParserProtocol
from pytest_bdd.compatibility.pytest import Config, Parser, get_config_root_path
from pytest_bdd.mimetypes import Mimetype
from pytest_bdd.model import Feature
from pytest_bdd.model.messages import Pickle
from pytest_bdd.parser import GherkinParser
from pytest_bdd.utils import PytestBDDIdGeneratorHandler, compose, make_python_name

Args = collections.namedtuple("Args", ["args", "kwargs"])


def add_options(parser: Parser):
    """Add pytest-bdd options."""
    group = parser.getgroup("bdd", "Scenario")
    group.addoption(
        "--disable-feature-autoload",
        action="store_false",
        dest="feature_autoload",
        default=True,
        help="Turn off feature files autoload",
    )
    parser.addini(
        "disable_feature_autoload",
        default=True,
        type="bool",
        help="Turn off feature files autoload",
    )


@attrs
class UrlScenarioLocator:
    url_paths = attrib()
    filter_ = attrib()
    encoding = attrib()
    features_base_url = attrib()
    mimetype = attrib()
    parser_type = attrib()
    parse_args = attrib()

    async def fetch(self, session: aiohttp.ClientSession, url):
        sslcontext = ssl.create_default_context(cafile=certifi.where())
        async with session.get(url, ssl=sslcontext) as response:
            return response.content_type, await response.text(encoding=self.encoding)

    async def fetch_all(self, urls):
        async with aiohttp.ClientSession() as session:
            return await asyncio.gather(*[self.fetch(session, url) for url in urls], return_exceptions=True)

    def resolve(self, config: Union[Config, PytestBDDIdGeneratorHandler]):
        urls = list(
            self.url_paths
            if self.features_base_url is None
            else map(partial(urljoin, self.features_base_url), self.url_paths)
        )
        if not urls:
            return
        loop = asyncio.new_event_loop()
        responses = loop.run_until_complete(self.fetch_all(urls))

        # Wait 250 ms for the underlying SSL connections to close
        loop.run_until_complete(asyncio.sleep(0.250))
        loop.close()

        hook_handler = cast(Config, config).hook
        encoding = self.encoding

        for url, response in zip(urls, responses):
            if isinstance(response, Exception):
                continue

            mimetype, feature_content = response

            if self.mimetype is not None:
                mimetype = self.mimetype

                if isinstance(mimetype, Mimetype):
                    mimetype = mimetype.value

            if self.parser_type is None:
                parser_type = hook_handler.pytest_bdd_get_parser(
                    config=config,
                    mimetype=mimetype,
                )
            else:
                parser_type = self.parser_type

            if parser_type is None:
                break

            parser = parser_type(id_generator=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator)

            try:
                filename = None
                with NamedTemporaryFile(mode="w", delete=False) as f:
                    filename = f.name
                    f.write(feature_content)

                feature = parser.parse(
                    config,
                    Path(filename),
                    url,
                    *self.parse_args.args,
                    **{**dict(encoding=encoding), **self.parse_args.kwargs},
                )

                for pickle in feature.pickles:
                    if self.filter_ is None or self.filter_(config, feature, pickle):  # type: ignore
                        yield feature, pickle
            finally:
                if filename is not None:
                    with suppress(Exception):
                        os.unlink(filename)


@attrs
class FileScenarioLocator:
    feature_paths = attrib(default=Factory(list))
    filter_: Optional[Callable[[Config, Feature, Pickle], Tuple[Feature, Pickle]]] = attrib(default=None)
    encoding = attrib(default="utf-8")
    features_base_dir: Optional[Union[str, Path]] = attrib(default=None)
    mimetype: Optional[str] = attrib(default=None)
    parser_type: Optional[Type[ParserProtocol]] = attrib(default=GherkinParser)
    parse_args: Args = attrib(default=Factory(lambda: Args((), {})))

    def _resolve_features_base_dir(self, config: Union[Config, PytestBDDIdGeneratorHandler]):
        try:
            if self.features_base_dir is None:
                features_base_dir = cast(Config, config).getini("bdd_features_base_dir")
            else:
                features_base_dir = self.features_base_dir
        except (ValueError, KeyError):
            features_base_dir = get_config_root_path(cast(Config, config))
        else:
            if callable(features_base_dir):
                features_base_dir = features_base_dir(config)

            features_base_dir = (get_config_root_path(cast(Config, config)) / Path(features_base_dir)).resolve()

        return features_base_dir

    def _gen_feature_paths(self, features_base_dir):
        for feature_pathlike in self.feature_paths:
            if isinstance(feature_pathlike, Path):
                feature_path = features_base_dir / feature_pathlike
                if feature_path.is_dir():
                    yield from filter(methodcaller("is_file"), feature_path.glob("**/*"))
                else:
                    yield feature_path
            else:
                try:
                    yield from filter(methodcaller("is_file"), features_base_dir.glob(str(feature_pathlike)))
                except IndexError:
                    yield from filter(methodcaller("is_file"), features_base_dir.glob("**/*"))

    @staticmethod
    def _build_file_uri(features_base_dir: Path, feature_path: Path):
        if feature_path.is_absolute():
            try:
                common_path = Path(commonpath([feature_path, features_base_dir]))
            except ValueError:
                rel_feature_path = feature_path
            else:
                sub_levels = len(features_base_dir.relative_to(common_path).parts)
                sub_path = reduce(truediv, [".."] * sub_levels, Path())
                rel_feature_path = sub_path / feature_path.relative_to(common_path)
        else:
            rel_feature_path = feature_path

        return "file:" + str(rel_feature_path.as_posix())

    def resolve(self, config: Union[Config, PytestBDDIdGeneratorHandler]):
        features_base_dir = self._resolve_features_base_dir(config)
        already_resolved_feature_paths = set()

        for feature_path in self._gen_feature_paths(features_base_dir=features_base_dir):
            feature_path_key = str(feature_path)
            if feature_path_key in already_resolved_feature_paths:
                break

            uri = self._build_file_uri(features_base_dir, feature_path)
            already_resolved_feature_paths.add(feature_path_key)
            hook_handler = cast(Config, config).hook
            encoding = self.encoding

            if self.parser_type is None:
                if self.mimetype is None:
                    mimetype = hook_handler.pytest_bdd_get_mimetype(config=config, path=feature_path)
                else:
                    mimetype = self.mimetype

                parser_type = hook_handler.pytest_bdd_get_parser(
                    config=config,
                    mimetype=mimetype,
                )
            else:
                parser_type = self.parser_type

            if parser_type is None:
                break

            parser = parser_type(id_generator=cast(PytestBDDIdGeneratorHandler, config).pytest_bdd_id_generator)

            feature = parser.parse(
                config,
                feature_path,
                uri,
                *self.parse_args.args,
                **{**dict(encoding=encoding), **self.parse_args.kwargs},
            )

            for pickle in feature.pickles:
                if self.filter_ is None or self.filter_(config, feature, pickle):  # type: ignore
                    yield feature, pickle


def get_python_name_generator(name: str) -> Iterable[str]:
    """Generate a sequence of suitable python names out of given arbitrary string name."""
    python_name = make_python_name(name)
    suffix = ""
    index = 0

    def get_name() -> str:
        return "_".join(filter(bool, ["test", python_name, suffix]))

    while True:
        yield get_name()
        index += 1
        suffix = f"{index}"


test_names = get_python_name_generator("")


class FeaturePathType(Enum):
    PATH = "path"
    URL = "url"
    UNDEFINED = "undefined"


def scenario(
    feature_name: Optional[Union[Path, str]] = None,
    scenario_name: Optional[str] = None,
    encoding: str = "utf-8",
    features_base_dir: Optional[Union[Path, str]] = None,
    features_base_url=None,
    features_path_type: Optional[Union[FeaturePathType, str]] = None,
    features_mimetype: Optional[Mimetype] = None,
    return_test_decorator=True,
    parser_type: Optional[Type[ParserProtocol]] = None,
    parse_args=Args((), {}),
    locators=(),
):
    """
    Scenario decorator.

    :param feature_name: Feature file name. Absolute or relative to the configured feature base path.
    :param scenario_name: Scenario name.
    :param encoding: Feature file encoding.
    :param features_base_dir: Feature base directory from where features will be searched
    :param features_base_url: Feature base url from where features will be loaded
    :param features_path_type: If feature path is not absolute helps to select if filepath or url will be used
    :param features_mimetype: Helps to select appropriate parser if non-standard file extension is used
    :param return_test_decorator; Return test decorator or generated test
    :param parser_type: Parser used to parse feature-like file
    :param parse_args: args consumed by parser during parsing
    :param locators: Feature locators to load Features; Could be custom
    """
    return scenarios(
        *([feature_name] if feature_name is not None else []),
        filter_=scenario_name,
        encoding=encoding,
        features_base_dir=features_base_dir,
        features_base_url=features_base_url,
        features_path_type=features_path_type,
        features_mimetype=features_mimetype,
        return_test_decorator=return_test_decorator,
        locators=locators,
        parser_type=parser_type,
        parse_args=parse_args,
    )


def scenarios(
    *feature_paths: Union[Path, str],
    filter_: Optional[Union[str, Callable]] = None,
    return_test_decorator=False,
    encoding: str = "utf-8",
    features_base_dir: Optional[Union[Path, str]] = None,
    features_base_url: Optional[str] = None,
    features_path_type: Optional[Union[FeaturePathType, str]] = None,
    features_mimetype: Optional[Mimetype] = None,
    parser_type: Optional[Type[ParserProtocol]] = None,
    parse_args=Args((), {}),
    locators=(),
):
    """
    Function to bind feature files to pytest runtime

    :param feature_paths: Features file names. Absolute or relative to the configured feature base path.
    :param filter_: Callable to filter scenarios
    :param encoding: Feature file encoding.
    :param features_base_dir: Feature base directory from where features will be searched
    :param features_base_url: Feature base url from where features will be loaded
    :param features_path_type: If feature path is not absolute helps to select if filepath or url will be used
    :param features_mimetype: Helps to select appropriate parser if non-standard file extension is used
    :param return_test_decorator; Return test decorator or generated test
    :param parser_type: Parser used to parse feature-like file
    :param parser_type: Parser used to parse feature-like file
    :param parse_args: args consumed by parser during parsing
    :param return_test_decorator; Return test decorator or generated test
    :param locators: Feature locators to load Features; Could be custom
    """

    decorator = compose(
        mark.pytest_bdd_scenario,
        mark.usefixtures("feature", "scenario"),
        mark.scenarios(
            *feature_paths,
            filter_=filter_,
            encoding=encoding,
            features_base_dir=features_base_dir,
            features_base_url=features_base_url,
            features_path_type=features_path_type,
            features_mimetype=features_mimetype,
            parser_type=parser_type,
            parse_args=parse_args,
            locators=locators,
        ),
    )

    if return_test_decorator:
        return decorator
    else:

        @decorator
        def test():
            ...

        test.__name__ = next(iter(test_names))

        return test



"""StepHandler parsers."""
# from __future__ import annotations

from abc import ABCMeta, abstractmethod
from functools import partial, singledispatchmethod
from itertools import filterfalse
from operator import attrgetter, contains, methodcaller
from re import Match
from re import Pattern as _RePattern
from re import compile as re_compile
from typing import Any, Dict, Iterable, Optional, Sequence, Union, cast

import parse as base_parse
import parse_type.cfparse as base_cfparse
from cucumber_expressions.argument import Argument as CucumberExpressionArgument
from cucumber_expressions.expression import CucumberExpression
from cucumber_expressions.parameter_type_registry import ParameterTypeRegistry
from cucumber_expressions.regular_expression import RegularExpression as CucumberRegularExpression

from pytest_bdd.compatibility.typing import Protocol, runtime_checkable
from pytest_bdd.model.messages import ExpressionType
from pytest_bdd.utils import StringableProtocol, stringify


class ParserBuildValueError(ValueError):
    ...


@runtime_checkable
class StepParserProtocol(Protocol):
    type: ExpressionType = ExpressionType.pytest_bdd_other_expression

    def parse_arguments(
        self, name: str, anonymous_group_names: Optional[Iterable[str]] = None
    ) -> Optional[Dict[str, Any]]:
        ...  # pragma: no cover

    def is_matching(self, name: str) -> bool:
        ...  # pragma: no cover

    def __str__(self) -> str:
        ...  # pragma: no cover


class StepParser(StepParserProtocol, metaclass=ABCMeta):
    """Parser of the individual step."""

    @abstractmethod
    def parse_arguments(
        self, name: str, anonymous_group_names: Optional[Iterable[str]] = None
    ) -> Optional[Dict[str, Any]]:
        """Get step arguments from the given step name.

        :return: `dict` of step arguments
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def is_matching(self, name: str) -> bool:
        """Match given name with the step name."""
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def __str__(self) -> str:
        """Match given name with the step name."""
        raise NotImplementedError()  # pragma: no cover

    @classmethod
    def build(cls, parserlike: Union[str, bytes, "StepParser", StepParserProtocol]) -> "StepParser":
        """Get parser by given name.

        :param parserlike: name of the step to parse

        :return: step parser object
        :rtype: StepParser
        """

        if isinstance(parserlike, StepParserProtocol):
            parser = cast(StepParser, parserlike)
        elif isinstance(parserlike, _RePattern):
            parser = re(parserlike)
        elif isinstance(parserlike, base_parse.Parser):
            parser = parse(parserlike)
        elif isinstance(parserlike, CucumberExpression):
            parser = cucumber_expression(parserlike)
        elif isinstance(parserlike, CucumberRegularExpression):
            parser = cucumber_regular_expression(parserlike)
        else:
            parser = heuristic(parserlike)

        return parser


class re(StepParser):
    """Regex step parser."""

    type = ExpressionType.pytest_bdd_regular_expression

    # https://bugs.python.org/issue45684
    @singledispatchmethod  # type:ignore[misc]
    def __init__(self, *args, **kwargs):
        raise NotImplementedError()  # pragma: no cover

    @__init__.register
    def _(self, pattern: str, *args: Any, **kwargs: Any) -> None:
        """Compile regex."""
        self.pattern = pattern
        self.regex = re_compile(self.pattern, *args, **kwargs)

    @__init__.register
    def _(self, pattern: _RePattern):
        """Compile regex."""
        self.pattern = pattern.pattern
        self.regex = pattern

    def parse_arguments(
        self,
        name,
        anonymous_group_names: Optional[Iterable[str]] = None,
    ):
        match = cast(Match, self.regex.match(name))  # Can't be None because is already matched
        group_dict = match.groupdict()
        if anonymous_group_names is not None:
            group_dict.update(
                zip(
                    anonymous_group_names,
                    map(
                        lambda span: name[slice(*span)],  # type: ignore[no-any-return] # https://github.com/python/mypy/issues/9590
                        filterfalse(
                            partial(contains, [*map(match.span, group_dict.keys())]),
                            map(match.span, range(1, len(match.groups()) + 1)),
                        ),
                    ),
                )
            )

        return group_dict

    def is_matching(self, name):
        return bool(self.regex.fullmatch(name))

    def __str__(self):
        return stringify(self.pattern)


class parse(StepParser):
    """parse step parser."""

    type = ExpressionType.pytest_bdd_parse_expression

    # https://bugs.python.org/issue45684
    @singledispatchmethod  # type:ignore[misc]
    def __init__(self, format, *args, **kwargs):
        if isinstance(format, (StringableProtocol, str, bytes)):
            self.__init_stringable__(format, *args, **kwargs)
        else:
            raise ParserBuildValueError(f"Unable build parser for format {format}")  # pragma: no cover

    def __init_stringable__(
        self, format: Union[StringableProtocol, str, bytes], *args: Any, builder=base_parse.compile, **kwargs: Any
    ) -> None:
        self.format = stringify(format)
        self.parser = builder(self.format, *args, **kwargs)

    @__init__.register
    def _(self, format: base_parse.Parser):
        self.format = format._format
        self.parser = format

    @classmethod
    def cfparse(cls, *args, **kwargs):
        kwargs.setdefault("builder", base_cfparse.Parser)
        return cls(*args, **kwargs)

    def parse_arguments(
        self, name: str, anonymous_group_names: Optional[Iterable[str]] = None
    ) -> Union[Dict[str, Any]]:
        match = self.parser.parse(name)
        group_dict = cast(dict, match.named)
        if anonymous_group_names is not None:
            group_dict.update(dict(zip(anonymous_group_names, match.fixed)))
        return group_dict

    def is_matching(self, name):
        try:
            return bool(self.parser.parse(name))
        except ValueError:
            return False

    def __str__(self):
        return str(self.format)


class cfparse(parse):
    """cfparse step parser."""

    type = ExpressionType.pytest_bdd_cfparse_expression

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("builder", base_cfparse.Parser)
        super().__init__(*args, **kwargs)


class string(StepParser):
    """Exact string step parser."""

    type = ExpressionType.pytest_bdd_string_expression

    def __init__(self, name: Union[StringableProtocol, str, bytes]) -> None:
        self.name = stringify(name)

    def parse_arguments(self, name: str, anonymous_group_names: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        """No parameters are available for simple string step.

        :return: `dict` of step arguments
        """
        return {}

    def is_matching(self, name: str) -> bool:
        """Match given name with the step name."""
        return bool(self.name == name)

    def __str__(self):
        return self.name


@runtime_checkable
class _CucumberExpressionProtocol(Protocol):
    def match(self, text: str) -> Optional[Sequence[CucumberExpressionArgument]]:
        ...  # pragma: no cover


class _CucumberExpression(StepParser):
    pattern: str
    expression: _CucumberExpressionProtocol

    def is_matching(self, name: str) -> bool:
        return bool(self.expression.match(name))

    def parse_arguments(
        self, name: str, anonymous_group_names: Optional[Iterable[str]] = None
    ) -> Optional[Dict[str, Any]]:
        return dict(zip(anonymous_group_names or [], map(attrgetter("value"), self.expression.match(name) or [])))

    def __str__(self):
        return str(self.pattern)


class cucumber_expression(_CucumberExpression):
    type = ExpressionType.cucumber_expression

    # https://bugs.python.org/issue45684
    @singledispatchmethod  # type:ignore[misc]
    def __init__(self, *args, **kwargs):
        raise NotImplementedError()  # pragma: no cover

    @__init__.register
    def _(self, expression: str, parameter_type_registry: ParameterTypeRegistry = ParameterTypeRegistry()):
        self.pattern = expression
        self.expression = CucumberExpression(expression, parameter_type_registry=parameter_type_registry)

    @__init__.register
    def _(self, expression: CucumberExpression):
        self.pattern = expression.expression
        self.expression = expression


class cucumber_regular_expression(_CucumberExpression):
    type = ExpressionType.regular_expression

    # https://bugs.python.org/issue45684
    @singledispatchmethod  # type:ignore[misc]
    def __init__(self, *args, **kwargs):
        raise NotImplementedError()  # pragma: no cover

    @__init__.register
    def _(self, expression: str, parameter_type_registry: ParameterTypeRegistry = ParameterTypeRegistry()):
        self.pattern = expression
        self.expression = CucumberRegularExpression(expression, parameter_type_registry=parameter_type_registry)

    @__init__.register
    def _(self, expression: CucumberRegularExpression):
        self.pattern = expression.expression_regexp.pattern
        self.expression = expression


class heuristic(StepParser):
    type = ExpressionType.pytest_bdd_heuristic_expression

    def __init__(self, format):
        if isinstance(format, (StringableProtocol, str, bytes)):
            format = stringify(format)
        self.format = format

        # Rework to exception groups after python 3.10 end of support
        e_cause = None
        try:
            self.string_parser = string(format)
        except Exception as e:
            e_cause = e
            self.string_parser = None

        try:
            self.cucumber_expression_parser = cucumber_expression(format)
        except Exception as e:
            e.__cause__, e_cause = e_cause, e
            self.cucumber_expression_parser = None

        try:
            self.cfparse_parser = cfparse(format)
        except Exception as e:
            e.__cause__, e_cause = e_cause, e
            self.cfparse_parser = None

        try:
            self.re_parser = re(format)
        except Exception as e:
            e.__cause__, e_cause = e_cause, e
            self.re_parser = None

        if not any(self.parser_by_priorities):
            raise ParserBuildValueError(f"Unable build parser for format {format}") from e_cause  # pragma: no cover

    @property
    def parser_by_priorities(self) -> Sequence[StepParser]:
        return [self.string_parser, self.cucumber_expression_parser, self.cfparse_parser, self.re_parser]

    def is_matching(self, name: str) -> bool:
        return any(map(methodcaller("is_matching", name), filter(bool, self.parser_by_priorities)))

    def parse_arguments(
        self, name: str, anonymous_group_names: Optional[Iterable[str]] = None
    ) -> Optional[Dict[str, Any]]:
        for parser in self.parser_by_priorities:
            if parser is not None and parser.is_matching(name):
                arguments = parser.parse_arguments(name, anonymous_group_names=anonymous_group_names)
                break
        else:
            arguments = None
        return arguments

    def __str__(self):
        return self.format



# generated by datamodel-codegen, manually updated

from enum import Enum
from typing import List, Optional, Sequence

from pydantic import BaseModel, Extra, Field


class ContentEncoding(Enum):
    identity = "IDENTITY"
    base64 = "BASE64"


class MediaType(Enum):
    text_x_cucumber_gherkin_plain = "text/x.cucumber.gherkin+plain"
    text_x_cucumber_gherkin_markdown = "text/x.cucumber.gherkin+markdown"


class Source(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    uri: str = Field(
        ...,
        description="*\n The [URI](https://en.wikipedia.org/wiki/Uniform_Resource_Identifier)\n of the source, typically a file path relative to the root directory",
    )
    data: str = Field(..., description="The contents of the file")
    media_type: MediaType = Field(
        ...,
        alias="mediaType",
        description="The media type of the file. Can be used to specify custom types, such as\n text/x.cucumber.gherkin+plain",
    )


class Location(BaseModel):
    class Config:
        extra = Extra.forbid

    line: int
    column: Optional[int] = None


class Comment(BaseModel):
    class Config:
        extra = Extra.forbid

    location: Location = Field(..., description="The location of the comment")
    text: str = Field(..., description="The text of the comment")


class DocString(BaseModel):
    class Config:
        extra = Extra.forbid

    location: Location
    media_type: Optional[str] = Field(None, alias="mediaType")
    content: str
    delimiter: str


class KeywordType(Enum):
    unknown = "Unknown"
    context = "Context"
    action = "Action"
    outcome = "Outcome"
    conjunction = "Conjunction"


class TableCell(BaseModel):
    class Config:
        extra = Extra.forbid

    location: Location = Field(..., description="The location of the cell")
    value: str = Field(..., description="The value of the cell")


class TableRow(BaseModel):
    class Config:
        extra = Extra.forbid

    location: Location = Field(..., description="The location of the first cell in the row")
    cells: Sequence[TableCell] = Field(..., description="Cells in the row")
    id: str


class Tag(BaseModel):
    class Config:
        extra = Extra.forbid

    location: Location = Field(..., description="Location of the tag")
    name: str = Field(..., description="The name of the tag (including the leading `@`)")
    id: str = Field(..., description="Unique ID to be able to reference the Tag from PickleTag")


class JavaMethod(BaseModel):
    class Config:
        extra = Extra.forbid

    class_name: str = Field(..., alias="className")
    method_name: str = Field(..., alias="methodName")
    method_parameter_types: Sequence[str] = Field(..., alias="methodParameterTypes")


class JavaStackTraceElement(BaseModel):
    class Config:
        extra = Extra.forbid

    class_name: str = Field(..., alias="className")
    file_name: str = Field(..., alias="fileName")
    method_name: str = Field(..., alias="methodName")


class Git(BaseModel):
    class Config:
        extra = Extra.forbid

    remote: str
    revision: str
    branch: Optional[str] = None
    tag: Optional[str] = None


class Product(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str = Field(..., description="The product name")
    version: Optional[str] = Field(None, description="The product version")


class ParameterType(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str = Field(..., description="The name is unique, so we don't need an id.")
    regular_expressions: List[str] = Field(..., alias="regularExpressions", min_items=1)
    prefer_for_regular_expression_match: bool = Field(..., alias="preferForRegularExpressionMatch")
    use_for_snippets: bool = Field(..., alias="useForSnippets")
    id: str


class PickleDocString(BaseModel):
    class Config:
        extra = Extra.forbid

    media_type: Optional[str] = Field(None, alias="mediaType")
    content: str


class Type(Enum):
    unknown = "Unknown"
    context = "Context"
    action = "Action"
    outcome = "Outcome"


class PickleTableCell(BaseModel):
    class Config:
        extra = Extra.forbid

    value: str


class PickleTableRow(BaseModel):
    class Config:
        extra = Extra.forbid

    cells: List[PickleTableCell] = Field(..., min_items=1)


class PickleTag(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str
    ast_node_id: str = Field(
        ...,
        alias="astNodeId",
        description="Points to the AST node this was created from",
    )


class ExpressionType(Enum):
    cucumber_expression = "CUCUMBER_EXPRESSION"
    regular_expression = "REGULAR_EXPRESSION"

    pytest_bdd_heuristic_expression = "PYTEST_BDD_HEURISTIC_EXPRESSION"
    pytest_bdd_string_expression = "PYTEST_BDD_STRING_EXPRESSION"
    pytest_bdd_regular_expression = "PYTEST_BDD_REGULAR_EXPRESSION"
    pytest_bdd_parse_expression = "PYTEST_BDD_PARSE_EXPRESSION"
    pytest_bdd_cfparse_expression = "PYTEST_BDD_CFPARSE_EXPRESSION"
    pytest_bdd_other_expression = "PYTEST_BDD_OTHER_EXPRESSION"


class StepDefinitionPattern(BaseModel):
    class Config:
        extra = Extra.forbid

    source: str
    type: ExpressionType


class Group(BaseModel):
    class Config:
        extra = Extra.forbid

    children: List["Group"]
    start: Optional[int] = None
    value: Optional[str] = None


class StepMatchArgument(BaseModel):
    class Config:
        extra = Extra.forbid

    group: Group = Field(
        ...,
        description="*\n Represents the outermost capture group of an argument. This message closely matches the\n `Group` class in the `cucumber-expressions` library.",
    )
    parameter_type_name: Optional[str] = Field(None, alias="parameterTypeName")


class StepMatchArgumentsList(BaseModel):
    class Config:
        extra = Extra.forbid

    step_match_arguments: List[StepMatchArgument] = Field(..., alias="stepMatchArguments")


class TestStep(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    hook_id: Optional[str] = Field(
        None,
        alias="hookId",
        description="Pointer to the `Hook` (if derived from a Hook)",
    )
    id: str
    pickle_step_id: Optional[str] = Field(
        None,
        alias="pickleStepId",
        description="Pointer to the `PickleStep` (if derived from a `PickleStep`)",
    )
    step_definition_ids: Optional[List[str]] = Field(
        None,
        alias="stepDefinitionIds",
        description="Pointer to all the matching `StepDefinition`s (if derived from a `PickleStep`)",
    )
    step_match_arguments_lists: Optional[List[StepMatchArgumentsList]] = Field(
        None,
        alias="stepMatchArgumentsLists",
        description="A list of list of StepMatchArgument (if derived from a `PickleStep`).\n Each element represents a matching step definition. A size of 0 means `UNDEFINED`,\n and a size of 2+ means `AMBIGUOUS`",
    )


class Timestamp(BaseModel):
    class Config:
        extra = Extra.forbid

    seconds: int = Field(
        ...,
        description="Represents seconds of UTC time since Unix epoch\n 1970-01-01T00:00:00Z. Must be from 0001-01-01T00:00:00Z to\n 9999-12-31T23:59:59Z inclusive.",
    )
    nanos: int = Field(
        ...,
        description="Non-negative fractions of a second at nanosecond resolution. Negative\n second values with fractions must still have non-negative nanos values\n that count forward in time. Must be from 0 to 999,999,999\n inclusive.",
    )


class TestCaseStarted(BaseModel):
    class Config:
        extra = Extra.forbid

    attempt: int = Field(
        ...,
        description="*\n The first attempt should have value 0, and for each retry the value\n should increase by 1.",
    )
    id: str = Field(
        ...,
        description="*\n Because a `TestCase` can be run multiple times (in case of a retry),\n we use this field to group messages relating to the same attempt.",
    )
    test_case_id: str = Field(..., alias="testCaseId")
    worker_id: Optional[str] = Field(
        None,
        alias="workerId",
        description="An identifier for the worker process running this test case, if test cases are being run in parallel. The identifier will be unique per worker, but no particular format is defined - it could be an index, uuid, machine name etc - and as such should be assumed that it's not human readable.",
    )
    timestamp: Timestamp


class TestRunFinished(BaseModel):
    class Config:
        extra = Extra.forbid

    message: Optional[str] = Field(
        None,
        description='Error message. Can be a stack trace from a failed `BeforeAll` or `AfterAll`.\n If there are undefined parameter types, the message is simply\n "The following parameter type(s() are not defined: xxx, yyy".\n The independent `UndefinedParameterType` messages can be used to generate\n snippets for those parameter types.',
    )
    success: bool = Field(
        ...,
        description="success = StrictModeEnabled ? (failed_count == 0 && ambiguous_count == 0 && undefined_count == 0 && pending_count == 0) : (failed_count == 0 && ambiguous_count == 0)",
    )
    timestamp: Timestamp = Field(..., description="Timestamp when the TestRun is finished")


class TestRunStarted(BaseModel):
    class Config:
        extra = Extra.forbid

    timestamp: Timestamp


class Status(Enum):
    unknown = "UNKNOWN"
    passed = "PASSED"
    skipped = "SKIPPED"
    pending = "PENDING"
    undefined = "UNDEFINED"
    ambiguous = "AMBIGUOUS"
    failed = "FAILED"


class Duration(BaseModel):
    class Config:
        extra = Extra.forbid

    seconds: int
    nanos: int = Field(
        ...,
        description="Non-negative fractions of a second at nanosecond resolution. Negative\n second values with fractions must still have non-negative nanos values\n that count forward in time. Must be from 0 to 999,999,999\n inclusive.",
    )


class TestStepStarted(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    test_case_started_id: str = Field(..., alias="testCaseStartedId")
    test_step_id: str = Field(..., alias="testStepId")
    timestamp: Timestamp


class UndefinedParameterType(BaseModel):
    class Config:
        extra = Extra.forbid

    expression: str
    name: str


class Attachment(BaseModel):
    class Config:
        extra = Extra.forbid

    body: str = Field(
        ...,
        description="*\n The body of the attachment. If `contentEncoding` is `IDENTITY`, the attachment\n is simply the string. If it's `BASE64`, the string should be Base64 decoded to\n obtain the attachment.",
    )
    content_encoding: ContentEncoding = Field(
        ...,
        alias="contentEncoding",
        description='*\n Whether to interpret `body` "as-is" (IDENTITY) or if it needs to be Base64-decoded (BASE64).\n\n Content encoding is *not* determined by the media type, but rather by the type\n of the object being attached:\n\n - string => IDENTITY\n - byte array => BASE64\n - stream => BASE64',
    )
    file_name: Optional[str] = Field(
        None,
        alias="fileName",
        description="*\n Suggested file name of the attachment. (Provided by the user as an argument to `attach`)",
    )
    media_type: str = Field(
        ...,
        alias="mediaType",
        description="*\n The media type of the data. This can be any valid\n [IANA Media Type](https://www.iana.org/assignments/media-types/media-types.xhtml)\n as well as Cucumber-specific media types such as `text/x.cucumber.gherkin+plain`\n and `text/x.cucumber.stacktrace+plain`",
    )
    source: Optional[Source] = None
    test_case_started_id: Optional[str] = Field(None, alias="testCaseStartedId")
    test_step_id: Optional[str] = Field(None, alias="testStepId")
    url: Optional[str] = Field(
        None,
        description="*\n A URL where the attachment can be retrieved. This field should not be set by Cucumber.\n It should be set by a program that reads a message stream and does the following for\n each Attachment message:\n\n - Writes the body (after base64 decoding if necessary) to a new file.\n - Sets `body` and `contentEncoding` to `null`\n - Writes out the new attachment message\n\n This will result in a smaller message stream, which can improve performance and\n reduce bandwidth of message consumers. It also makes it easier to process and download attachments\n separately from reports.",
    )


class DataTable(BaseModel):
    class Config:
        extra = Extra.forbid

    location: Location
    rows: List[TableRow]


class Examples(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    location: Location = Field(..., description="The location of the `Examples` keyword")
    tags: List[Tag]
    keyword: str
    name: str
    description: str
    table_header: Optional[TableRow] = Field(None, alias="tableHeader")
    table_body: List[TableRow] = Field(..., alias="tableBody")
    id: str


class Step(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    location: Location = Field(..., description="The location of the steps' `keyword`")
    keyword: str = Field(..., description="The actual keyword as it appeared in the source.")
    keyword_type: Optional[KeywordType] = Field(
        None,
        alias="keywordType",
        description="The test phase signalled by the keyword: Context definition (Given), Action performance (When), Outcome assertion (Then). Other keywords signal Continuation (And and But) from a prior keyword. Please note that all translations which a dialect maps to multiple keywords (`*` is in this category for all dialects), map to 'Unknown'.",
    )
    text: str
    doc_string: Optional[DocString] = Field(None, alias="docString")
    data_table: Optional[DataTable] = Field(None, alias="dataTable")
    id: str = Field(..., description="Unique ID to be able to reference the Step from PickleStep")


class SourceReference(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    uri: Optional[str] = None
    java_method: Optional[JavaMethod] = Field(None, alias="javaMethod")
    java_stack_trace_element: Optional[JavaStackTraceElement] = Field(None, alias="javaStackTraceElement")
    location: Optional[Location] = None


class Ci(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str = Field(..., description='Name of the CI product, e.g. "Jenkins", "CircleCI" etc.')
    url: Optional[str] = Field(None, description="Link to the build")
    build_number: Optional[str] = Field(
        None,
        alias="buildNumber",
        description="The build number. Some CI servers use non-numeric build numbers, which is why this is a string",
    )
    git: Optional[Git] = None


class ParseError(BaseModel):
    class Config:
        extra = Extra.forbid

    source: SourceReference
    message: str


class PickleTable(BaseModel):
    class Config:
        extra = Extra.forbid

    rows: List[PickleTableRow]


class StepDefinition(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    id: str
    pattern: StepDefinitionPattern
    source_reference: SourceReference = Field(..., alias="sourceReference")


class TestCase(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    id: str
    pickle_id: str = Field(
        ...,
        alias="pickleId",
        description="The ID of the `Pickle` this `TestCase` is derived from.",
    )
    test_steps: List[TestStep] = Field(..., alias="testSteps")


class TestCaseFinished(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    test_case_started_id: str = Field(..., alias="testCaseStartedId")
    timestamp: Timestamp
    will_be_retried: bool = Field(..., alias="willBeRetried")


class TestStepResult(BaseModel):
    class Config:
        extra = Extra.forbid

    duration: Duration
    message: Optional[str] = None
    status: Status


class Background(BaseModel):
    class Config:
        extra = Extra.forbid

    location: Location = Field(..., description="The location of the `Background` keyword")
    keyword: str
    name: str
    description: str
    steps: List[Step]
    id: str


class Scenario(BaseModel):
    class Config:
        extra = Extra.forbid

    location: Location = Field(..., description="The location of the `Scenario` keyword")
    tags: List[Tag]
    keyword: str
    name: str
    description: str
    steps: List[Step]
    examples: List[Examples]
    id: str


class Hook(BaseModel):
    class Config:
        extra = Extra.forbid

    id: str
    name: Optional[str] = None
    source_reference: SourceReference = Field(..., alias="sourceReference")
    tag_expression: Optional[str] = Field(None, alias="tagExpression")


class Meta(BaseModel):
    class Config:
        extra = Extra.forbid

    protocol_version: str = Field(
        ...,
        alias="protocolVersion",
        description="*\n The [SEMVER](https://semver.org/) version number of the protocol",
    )
    implementation: Product = Field(
        ...,
        description="SpecFlow, Cucumber-JVM, Cucumber.js, Cucumber-Ruby, Behat etc.",
    )
    runtime: Product = Field(..., description="Java, Ruby, Node.js etc")
    os: Product = Field(..., description="Windows, Linux, MacOS etc")
    cpu: Product = Field(..., description="386, arm, amd64 etc")
    ci: Optional[Ci] = None


class PickleStepArgument(BaseModel):
    class Config:
        extra = Extra.forbid

    doc_string: Optional[PickleDocString] = Field(None, alias="docString")
    data_table: Optional[PickleTable] = Field(None, alias="dataTable")


class TestStepFinished(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    test_case_started_id: str = Field(..., alias="testCaseStartedId")
    test_step_id: str = Field(..., alias="testStepId")
    test_step_result: TestStepResult = Field(..., alias="testStepResult")
    timestamp: Timestamp


class RuleChild(BaseModel):
    class Config:
        extra = Extra.forbid

    background: Optional[Background] = None
    scenario: Optional[Scenario] = None


class PickleStep(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    argument: Optional[PickleStepArgument] = None
    ast_node_ids: List[str] = Field(
        ...,
        alias="astNodeIds",
        description="References the IDs of the source of the step. For Gherkin, this can be\n the ID of a Step, and possibly also the ID of a TableRow",
    )
    id: str = Field(..., description="A unique ID for the PickleStep")
    type: Optional[Type] = Field(
        None,
        description="The context in which the step was specified: context (Given), action (When) or outcome (Then).\n\nNote that the keywords `But` and `And` inherit their meaning from prior steps and the `*` 'keyword' doesn't have specific meaning (hence Unknown)",
    )
    text: str


class Rule(BaseModel):
    class Config:
        extra = Extra.forbid

    location: Location = Field(..., description="The location of the `Rule` keyword")
    tags: List[Tag] = Field(..., description="All the tags placed above the `Rule` keyword")
    keyword: str
    name: str
    description: str
    children: List[RuleChild]
    id: str


class Pickle(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    id: str = Field(..., description="*\n A unique id for the pickle")
    uri: str = Field(..., description="The uri of the source file")
    name: str = Field(..., description="The name of the pickle")
    language: str = Field(..., description="The language of the pickle")
    steps: List[PickleStep] = Field(..., description="One or more steps")
    tags: List[PickleTag] = Field(
        ...,
        description="*\n One or more tags. If this pickle is constructed from a Gherkin document,\n It includes inherited tags from the `Feature` as well.",
    )
    ast_node_ids: List[str] = Field(
        ...,
        alias="astNodeIds",
        description="*\n Points to the AST node locations of the pickle. The last one represents the unique\n id of the pickle. A pickle constructed from `Examples` will have the first\n id originating from the `Scenario` AST node, and the second from the `TableRow` AST node.",
        min_items=1,
    )


class FeatureChild(BaseModel):
    class Config:
        extra = Extra.forbid

    rule: Optional[Rule] = None
    background: Optional[Background] = None
    scenario: Optional[Scenario] = None


class Feature(BaseModel):
    class Config:
        extra = Extra.forbid

    location: Location = Field(..., description="The location of the `Feature` keyword")
    tags: List[Tag] = Field(..., description="All the tags placed above the `Feature` keyword")
    language: str = Field(
        ...,
        description="The [ISO 639-1](https://en.wikipedia.org/wiki/ISO_639-1) language code of the Gherkin document",
    )
    keyword: Optional[str] = Field(
        None,
        description="The text of the `Feature` keyword (in the language specified by `language`)",
    )
    name: str = Field(..., description="The name of the feature (the text following the `keyword`)")
    description: str = Field(
        ...,
        description="The line(s) underneath the line with the `keyword` that are used as description",
    )
    children: List[FeatureChild] = Field(..., description="Zero or more children")


class GherkinDocument(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    uri: Optional[str] = Field(
        None,
        description="*\n The [URI](https://en.wikipedia.org/wiki/Uniform_Resource_Identifier)\n of the source, typically a file path relative to the root directory",
    )
    feature: Optional[Feature] = None
    comments: List[Comment] = Field(..., description="All the comments in the Gherkin document")


class Message(BaseModel):
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True

    attachment: Optional[Attachment] = None
    gherkin_document: Optional[GherkinDocument] = Field(None, alias="gherkinDocument")
    hook: Optional[Hook] = None
    meta: Optional[Meta] = None
    parameter_type: Optional[ParameterType] = Field(None, alias="parameterType")
    parse_error: Optional[ParseError] = Field(None, alias="parseError")
    pickle: Optional[Pickle] = None
    source: Optional[Source] = None
    step_definition: Optional[StepDefinition] = Field(None, alias="stepDefinition")
    test_case: Optional[TestCase] = Field(None, alias="testCase")
    test_case_finished: Optional[TestCaseFinished] = Field(None, alias="testCaseFinished")
    test_case_started: Optional[TestCaseStarted] = Field(None, alias="testCaseStarted")
    test_run_finished: Optional[TestRunFinished] = Field(None, alias="testRunFinished")
    test_run_started: Optional[TestRunStarted] = Field(None, alias="testRunStarted")
    test_step_finished: Optional[TestStepFinished] = Field(None, alias="testStepFinished")
    test_step_started: Optional[TestStepStarted] = Field(None, alias="testStepStarted")
    undefined_parameter_type: Optional[UndefinedParameterType] = Field(None, alias="undefinedParameterType")


Group.update_forward_refs()



from importlib_resources import *
from importlib_resources import as_file, files

__all__ = [
    "Package",
    "Resource",
    "ResourceReader",
    "as_file",
    "contents",
    "files",
    "is_resource",
    "open_binary",
    "open_text",
    "path",
    "read_binary",
    "read_text",
]



from itertools import filterfalse
from json import loads as json_loads
from operator import attrgetter
from typing import Any, Union, cast

from attr import attrib, attrs
from gherkin.pickles.compiler import Compiler

from pytest_bdd.model.gherkin_document import Feature as GherkinDocumentFeature
from pytest_bdd.model.messages import (
    DataTable,
    DocString,
    Examples,
    Feature,
    FeatureChild,
    GherkinDocument,
    KeywordType,
    Location,
    Scenario,
    Step,
    TableCell,
    TableRow,
    Tag,
    Type,
)
from pytest_bdd.struct_bdd.model import Join as StructJoin
from pytest_bdd.struct_bdd.model import StepPrototype as StructStep
from pytest_bdd.struct_bdd.model import Table as StructTable


@attrs
class _ASTBuilder:
    model: Any

    def build(self, *args, **kwargs):  # pragma: no cover
        raise NotImplementedError


@attrs
class GherkinDocumentBuilder(_ASTBuilder):
    model: StructStep = attrib()

    def build(self, id_generator):
        return GherkinDocument(
            comments=[], uri=None, feature=StepToFeatureASTBuilder(self.model).build(id_generator=id_generator)
        )

    def build_feature(self, filename, uri, id_generator):
        gherkin_document = self.build(id_generator=id_generator)
        gherkin_document.uri = uri

        gherkin_document_serialized = gherkin_document.json(by_alias=True, exclude_none=True)

        scenarios_data = Compiler().compile(json_loads(gherkin_document_serialized))
        pickles = GherkinDocumentFeature.load_pickles(scenarios_data)

        feature = GherkinDocumentFeature(  # type: ignore[call-arg]
            gherkin_document=gherkin_document,
            uri=uri,
            pickles=pickles,
            filename=filename,
        )

        feature.fill_registry()

        return feature


@attrs
class StepToFeatureASTBuilder(_ASTBuilder):
    model: StructStep = attrib()

    def build(self, id_generator):
        return Feature(
            children=self._build_children(id_generator=id_generator),
            description=self.model.description or "",
            language="EN",
            location=Location(column=0, line=0),
            tags=[],
            name=self.model.name or "",
            keyword="Feature",
        )

    def _build_children(self, id_generator):
        def _():
            for route in self.model.routes:
                if route.steps:

                    def steps_gen(steps):
                        previous_step_keyword_type = None
                        for step in steps:
                            step_keyword_type = (
                                previous_step_keyword_type
                                if step.keyword_type is KeywordType.conjunction
                                else step.keyword_type
                            )
                            yield Step(
                                id=next(id_generator),
                                keyword=step.type if isinstance(step.type, str) else cast(Type, step.type).value,
                                location=Location(column=0, line=0),
                                text=step.action,
                                keyword_type=step_keyword_type.value,
                                **(
                                    (
                                        lambda rows: dict(
                                            data_table=DataTable(
                                                rows=rows,
                                                location=Location(column=0, line=0),  # type: ignore[call-arg]
                                            )  # type: ignore[call-arg]
                                        )
                                        if rows
                                        else {}
                                    )(
                                        [
                                            *filterfalse(
                                                lambda row: row is None,
                                                map(
                                                    lambda row_values: (
                                                        (
                                                            lambda cells: TableRow(
                                                                id=next(id_generator),
                                                                location=Location(column=0, line=0),  # type: ignore[call-arg]
                                                                cells=cells,
                                                            )  # type: ignore[call-arg]
                                                            if cells
                                                            else None
                                                        )(
                                                            [
                                                                *map(
                                                                    lambda parameter: TableCell(
                                                                        location=Location(column=0, line=0),
                                                                        value=parameter,
                                                                    ),
                                                                    row_values,
                                                                )
                                                            ]
                                                        )
                                                    ),
                                                    StructJoin(tables=step.data).rowed_values,
                                                ),
                                            )
                                        ]
                                    )
                                ),
                                **(
                                    dict(
                                        doc_string=DocString(
                                            content=step.description,
                                            delimiter="\n",
                                            location=Location(column=0, line=0),
                                        )
                                    )
                                    if step.description
                                    else dict()
                                ),
                            )
                            previous_step_keyword_type = step_keyword_type

                    steps = [*steps_gen(filter(lambda step: step.action is not None, route.steps))]

                    yield FeatureChild(
                        scenario=Scenario(
                            description=route.steps[0].description or "",
                            examples=[ExampleASTBuilder(route.example_table).build(id_generator=id_generator)]
                            if route.example_table.values
                            else [],
                            id=next(id_generator),
                            keyword="Scenario",
                            location=Location(column=0, line=0),
                            name=next(filter(bool, map(attrgetter("name"), reversed(route.steps))), ""),
                            tags=[
                                *map(
                                    lambda tag_name: Tag(
                                        id=next(id_generator),
                                        location=Location(column=0, line=0),
                                        name=tag_name,
                                    ),
                                    route.tags,
                                )
                            ],
                            steps=steps,
                        )
                    )

        return list(_())


@attrs
class ExampleASTBuilder(_ASTBuilder):
    model: Union[StructJoin, StructTable] = attrib()

    def build(self, id_generator):
        return Examples(
            description=self.model.description,
            id=next(id_generator),
            keyword="Examples",
            location=Location(column=0, line=0),
            name=self.model.name,
            table_body=[
                *map(
                    lambda row_values: TableRow(
                        id=next(id_generator),
                        location=Location(column=0, line=0),
                        cells=[
                            *map(
                                lambda parameter: TableCell(
                                    location=Location(column=0, line=0),
                                    value=parameter,
                                ),
                                row_values,
                            )
                        ],
                    ),
                    self.model.rowed_values,
                )
            ],
            tags=[
                *map(
                    lambda tag_name: Tag(
                        id=next(id_generator),
                        location=Location(column=0, line=0),
                        name=tag_name,
                    ),
                    self.model.tags,
                )
            ],
            table_header=TableRow(
                id=next(id_generator),
                location=Location(column=0, line=0),
                cells=[
                    *map(
                        lambda parameter: TableCell(
                            location=Location(column=0, line=0),
                            value=parameter,
                        ),
                        self.model.parameters,
                    )
                ],
            ),
        )



from enum import Enum


class Mimetype(Enum):
    gherkin_plain = "text/x.cucumber.gherkin+plain"
    markdown = "text/x.cucumber.gherkin+markdown"
    struct_bdd_yaml = "application/x.struct_bdd+yaml"
    struct_bdd_hocon = "application/x.struct_bdd+hocon"
    struct_bdd_json5 = "application/x.struct_bdd+json5"
    struct_bdd_json = "application/x.struct_bdd+json"
    struct_bdd_hjson = "application/x.struct_bdd+hjson"
    struct_bdd_toml = "application/x.struct_bdd+toml"
